/*
 * Copyright (c) 1997-2001, 2003 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden). 
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 */

#include "push_locl.h"
RCSID("$Id: push.c,v 1.47 2003/04/04 02:10:17 assar Exp $");

#undef KRB4

#ifdef KRB4
static int use_v4 = -1;
#endif

#ifdef KRB5
static int use_sasl = -1;
static int use_v5 = -1;
static krb5_context context;
#endif

static char *kport_str;
static char *sport_str;
static int verbose_level;
static int do_fork;
static int do_leave;
static int do_version;
static int do_help;
static int do_from;
static int do_count;
static int do_encrypt = 1;
static char *header_str;

#undef warn
#undef warnx

#define warn printf
#define warnx printf

struct getargs args[] = {
#ifdef KRB4
    { "krb4",	'4', arg_flag,		&use_v4,	"Use Kerberos V4",
      NULL },
#endif    
#ifdef KRB5
    { "krb5",	'5', arg_flag,		&use_v5,	"Use Kerberos V5",
      NULL },
    { "sasl",	's', arg_flag,		&use_sasl,	"Use SASL AUTH",
      NULL },
#endif
    { "verbose",'v', arg_counter,	&verbose_level,	"Verbose",
      NULL },
    { "fork",	'f', arg_flag,		&do_fork,	"Fork deleting proc",
      NULL },
    { "leave",	'l', arg_flag,		&do_leave,	"Leave mail on server",
      NULL },
    { "kport",	'p', arg_string,	&kport_str,	"Use this port for KPOP",
      "number-or-service" },
    { "sport",	's', arg_string,	&sport_str,	"Use this port for SASL POP",
      "number-or-service" },
    { "from",	 0,  arg_flag,		&do_from,	"Behave like from",
      NULL },
    { "headers", 0,  arg_string,	&header_str,	"Headers to print", NULL },
    { "count", 'c',  arg_flag,		&do_count,	"Print number of messages", NULL},
    { "encrypt", 0,  arg_negative_flag,	&do_encrypt,	"Don't encrypt when using SASL"}, 
    { "version", 0,  arg_flag,		&do_version,	"Print version",
      NULL },
    { "help",	 0,  arg_flag,		&do_help,	NULL,
      NULL }

};

char *strndup(char *src, size_t len)
{
	char *r = malloc(len+1);
	if (r) {
		strlcpy(r, src, len);
		r[len] = 0;
		}
	return r;
}

ssize_t
writev(int d, const struct iovec *iov, int iovcnt)
{
	register int i;
 
	for (i = 0; i < iovcnt; i++) {
		if (write(d, iov[i].iov_base, iov[i].iov_len) != iov[i].iov_len) {
			fprintf(stderr, "Skrivfel\n");
			exit(1);
			}
		}
	return 0;
}

static void
usage (int ret)
{
    arg_printusage (args,
		    sizeof(args) / sizeof(args[0]),
		    NULL,
		    "[[{po:username[@hostname] | hostname[:username]}] ...] "
		    "filename");
    exit (ret);
}

static int
do_connect (const char *hostname, int port, int pf, int nodelay, void *ctx)
{
    struct addrinfo *ai, *a;
    struct addrinfo hints;
    int error;
    int s = -1;
    char portstr[NI_MAXSERV];

    memset (&hints, 0, sizeof(hints));
    hints.ai_family = pf;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_protocol = IPPROTO_TCP;

    snprintf (portstr, sizeof(portstr), "%u", ntohs(port));

    error = getaddrinfo (hostname, portstr, &hints, &ai);
    if (error)
	errx (1, "getaddrinfo(%s): %s", hostname, gai_strerror(error));

    for (a = ai; a != NULL; a = a->ai_next) {
	s = socket (a->ai_family, a->ai_socktype, a->ai_protocol);
	if (s < 0)
	    continue;
	if (connect (s, a->ai_addr, a->ai_addrlen) < 0) {
	    warn ("connect(%s)", hostname);
 	    close (s);
 	    continue;
	}
	break;
    }
    freeaddrinfo (ai);
    if (a == NULL) {
	warnx ("failed to contact %s", hostname);
	return -1;
    }

    if(setsockopt(s, IPPROTO_TCP, TCP_NODELAY,
		  (void *)&nodelay, sizeof(nodelay)) < 0)
	err (1, "setsockopt TCP_NODELAY");

#ifdef KRB5
    if (ctx) {
	int ret;

	ret = tsasl_prot_init(s, (struct tsasl_prot **)ctx);
	if (ret)
	    errx(1, "tsasl_prot_init: failed with %d", ret);
    }
#endif

    return s;
}

typedef enum { INIT, GREET, USER, PASS, STAT, RETR, TOP, 
	       DELE, XDELE, QUIT } pop_state;

static char *pop_state_string[] = {
    "INIT", "GREET", "USER", "PASS", "STAT", "RETR", "TOP",
    "DELE", "XDELE", "QUIT" 
};

#define PUSH_BUFSIZ 65536

#define STEP 16

struct write_state {
    struct iovec *iovecs;
    size_t niovecs, maxiovecs, allociovecs;
    int fd;
};

static void
write_state_init (struct write_state *w, int fd)
{
#ifdef UIO_MAXIOV
    w->maxiovecs = UIO_MAXIOV;
#else
    w->maxiovecs = 16;
#endif
    w->allociovecs = min(STEP, w->maxiovecs);
    w->niovecs = 0;
    w->iovecs = malloc(w->allociovecs * sizeof(*w->iovecs));
    w->fd = fd;
}

static void
write_state_add (struct write_state *w, void *v, size_t len)
{
    if(w->niovecs == w->allociovecs) {				
	if(w->niovecs == w->maxiovecs) {				
	    if(writev (w->fd, w->iovecs, w->niovecs) < 0)		
		err(1, "writev");				
	    w->niovecs = 0;					
	} else {						
	    w->allociovecs = min(w->allociovecs + STEP, w->maxiovecs);	
	    w->iovecs = realloc (w->iovecs,				
				  w->allociovecs * sizeof(*w->iovecs));	
	}							
    }								
    w->iovecs[w->niovecs].iov_base = v;				
    w->iovecs[w->niovecs].iov_len  = len;				
    ++w->niovecs;							
}

static void
write_state_flush (struct write_state *w)
{
    if (w->niovecs) {
	if (writev (w->fd, w->iovecs, w->niovecs) < 0)
	    err (1, "writev");
	w->niovecs = 0;
    }
}

static void
write_state_destroy (struct write_state *w)
{
    free (w->iovecs);
}

#if defined(KRB4) && !defined(KRB5)
static ssize_t
push_write(int fd, void *ctx, void *data, size_t len)
{
    return net_write(fd, data, len);
}

static ssize_t
push_read(int fd, void *ctx, void *data, size_t len)
{
    ssize_t ret;
    ret = net_read(fd, data, len);
    if (ret == 0)
	return -1;
    if (ret < 0)
	return -2;
    return ret;
}
#else
static ssize_t
push_write(int fd, void *ctx, void *data, size_t len)
{
    struct tsasl_prot *prot = ctx;
    if (verbose_level > 1)
	printf("%.*s", (int)len, (char*)data);
    return tsasl_write(prot, TSASL_PROT_FLUSH, data, len);
}

static ssize_t
push_read(int fd, void *ctx, void *data, size_t len)
{
    struct tsasl_prot *prot = ctx;
    return tsasl_read(prot, 0, data, len);
}
#endif

static int
doit(int s,
     void *ctx,
     const char *host,
     const char *user,
     const char *outfilename,
     const char *header_str,
     int leavep,
     int verbose,
     int forkp,
     pop_state start_state,
     char *auth_in_data,
     size_t auth_in_size)
{
    int ret;
    char out_buf[PUSH_BUFSIZ];
    int out_len = 0;
    char in_buf[PUSH_BUFSIZ + 1];	/* sentinel */
    size_t in_len = 0;
    char *in_ptr = in_buf;
    pop_state state = start_state;
    unsigned count, bytes;
    unsigned asked_for = 0, retrieved = 0, asked_deleted = 0, deleted = 0;
    unsigned sent_xdele = 0;
    int out_fd;
    char from_line[128];
    size_t from_line_length;
    time_t now;
    struct write_state write_state;
    int numheaders = 1;
    char **headers = NULL;
    int i;
    char *tmp = NULL;

    if (auth_in_size > sizeof(in_buf) - 1)
	errx(1, "doit: got in buffer from auth system that was larger then doit can handle");
    memcpy(in_ptr, auth_in_data, auth_in_size);
    in_len = auth_in_size;

    if (do_from) {
	char *tmp2;

	tmp2 = tmp = strdup(header_str);

	out_fd = -1;
	if (verbose)
	    fprintf (stderr, "%s@%s\n", user, host);
	while (*tmp != '\0') {
	    tmp = strchr(tmp, ',');
	    if (tmp == NULL)
		break;
	    tmp++;
	    numheaders++;
	}

	headers = malloc(sizeof(char *) * (numheaders + 1));
	for (i = 0; i < numheaders; i++) {
	    headers[i] = strtok_r(tmp2, ",", &tmp2);
	}
	headers[numheaders] = NULL;
    } else {
	out_fd = open(outfilename, O_WRONLY | O_APPEND | O_CREAT, 0666);
	if (out_fd < 0)
	    err (1, "open %s", outfilename);
	if (verbose)
	    fprintf (stderr, "%s@%s -> %s\n", user, host, outfilename);
    }

    now = time(NULL);
    from_line_length = snprintf (from_line, sizeof(from_line),
				 "From %s %s", "push", ctime(&now));

    if (state == INIT) {
	out_len = snprintf (out_buf, sizeof(out_buf),
			    "USER %s\r\nPASS hej\r\nSTAT\r\n",
			    user);
    } else if (state == PASS) {
	out_len = snprintf (out_buf, sizeof(out_buf),
			    "STAT\r\n");
    } else
	errx(1, "foobarkakakb");
    if (out_len < 0)
	errx (1, "snprintf failed");
    if (push_write (s, ctx, out_buf, out_len) != out_len)
	err (1, "write");
    if (verbose > 1)
	fprintf (stderr, "%s", out_buf);

    if (!do_from)
	write_state_init (&write_state, out_fd);

    while(state != QUIT) {
	fd_set readset, writeset;

	FD_ZERO(&readset);
	FD_ZERO(&writeset);
#if 0
	if (s >= FD_SETSIZE)
	    errx (1, "fd too large");
#endif
	FD_SET(s,&readset);

	if (verbose > 1)
	    fprintf (stderr, "state: %s count: %d asked_for: %d retrieved: %d asked_deleted: %d\n",
		     pop_state_string[state], 
		     count, asked_for, retrieved, asked_deleted);

	if (((state == STAT || state == RETR || state == TOP)
	     && asked_for < count)
	    || (state == XDELE && !sent_xdele)
	    || (state == DELE && asked_deleted < count))
	    FD_SET(s,&writeset);
	ret = select (s + 1, &readset, &writeset, NULL, NULL);
	if (ret < 0) {
	    if (errno == EAGAIN)
		continue;
	    else
		err (1, "select");
	}
	
	if (FD_ISSET(s, &readset)) {
	    char *beg, *p;
	    size_t rem;
	    int blank_line = 0;
	    
	    ret = push_read (s, ctx, in_ptr, sizeof(in_buf) - in_len - 1);
	    if (ret == -1)
		errx (1, "EOF during read");
	    else if (ret < 0)
		errx (1, "read: %d", ret);
	    
	    in_len += ret;
	    in_ptr += ret;
	    *in_ptr = '\0';
	    
	    beg = in_buf;
	    rem = in_len;
	    while((rem > 1) && (p = strstr(beg, "\r\n")) != NULL) {
		if (state == TOP) {
		    char *copy = beg;

		    for (i = 0; i < numheaders; i++) {
			size_t len;

			len = min(p - copy + 1, strlen(headers[i]));
			if (strncasecmp(copy, headers[i], len) == 0) {
			    fprintf (stdout, "%.*s\n", (int)(p - copy), copy);
			}
		    }
		    if (beg[0] == '.' && beg[1] == '\r' && beg[2] == '\n') {
			if (numheaders > 1)
			    fprintf (stdout, "\n");
			state = STAT;
			if (++retrieved == count) {
			    state = QUIT;
			    push_write (s, ctx, "QUIT\r\n", 6);
			    if (verbose > 1)
				fprintf (stderr, "QUIT\r\n");
			}
		    }
		    rem -= p - beg + 2;
		    beg = p + 2;
		} else if (state == RETR) {
		    char *copy = beg;
		    if (beg[0] == '.') {
			if (beg[1] == '\r' && beg[2] == '\n') {
			    if(!blank_line)
				write_state_add(&write_state, "\n", 1);
			    state = STAT;
			    rem -= p - beg + 2;
			    beg = p + 2;
			    if (++retrieved == count) {
				write_state_flush (&write_state);
//				if (fsync (out_fd) < 0)
//				    err (1, "fsync");
				close(out_fd);
				if (leavep) {
				    state = QUIT;
				    push_write (s, ctx, "QUIT\r\n", 6);
				    if (verbose > 1)
					fprintf (stderr, "QUIT\r\n");
				} else {
				    if (forkp) {
					    fprintf(stderr, "NO FORK\n");
					    exit(1);
					    }

				    state = XDELE;
				    if (verbose)
					fprintf (stderr, "deleting... ");
				}
			    }
			    continue;
			} else
			    ++copy;
		    }
		    *p = '\n';
		    if(blank_line && 
		       strncmp(copy, "From ", min(p - copy + 1, 5)) == 0)
			write_state_add(&write_state, ">", 1);
		    write_state_add(&write_state, copy, p - copy + 1);
		    blank_line = (*copy == '\n');
		    rem -= p - beg + 2;
		    beg = p + 2;
		} else if (rem >= 3 && strncmp (beg, "+OK", 3) == 0) {
		    if (state == STAT) {
			if (!do_from)
			    write_state_add(&write_state,
					    from_line, from_line_length);
			blank_line = 0;
			if (do_from) 
			    state = TOP;
			else
			    state = RETR;
		    } else if (state == XDELE) {
			state = QUIT;
			push_write (s, ctx, "QUIT\r\n", 6);
			if (verbose > 1)
			    fprintf (stderr, "QUIT\r\n");
			break;
		    } else if (state == DELE) {
			if (++deleted == count) {
			    state = QUIT;
			    push_write (s, ctx, "QUIT\r\n", 6);
			    if (verbose > 1)
				fprintf (stderr, "QUIT\r\n");
			    break;
			}
		    } else if (++state == STAT) {
			if(sscanf (beg + 4, "%u %u", &count, &bytes) != 2)
			    errx(1, "Bad STAT-line: %.*s", (int)(p - beg), beg);
			if (verbose) {
			    fprintf (stderr, "%u message(s) (%u bytes). "
				     "fetching... ",
				     count, bytes);
			    if (do_from)
				fprintf (stderr, "\n");
			} else if (do_count) {
			    fprintf (stderr, "%u message(s) (%u bytes).\n",
				     count, bytes);
			}
			if (count == 0) {
			    state = QUIT;
			    push_write (s, ctx, "QUIT\r\n", 6);
			    if (verbose > 1)
				fprintf (stderr, "QUIT\r\n");
			    break;
			}
		    }

		    rem -= p - beg + 2;
		    beg = p + 2;
		} else {
		    if(state == XDELE) {
			state = DELE;
			rem -= p - beg + 2;
			beg = p + 2;
		    } else
			errx (1, "Bad response: %.*s", (int)(p - beg), beg);
		}
	    }
	    if (!do_from)
		write_state_flush (&write_state);

	    memmove (in_buf, beg, rem);
	    in_len = rem;
	    in_ptr = in_buf + rem;
	}
	if (FD_ISSET(s, &writeset)) {
	    if ((state == STAT && !do_from) || state == RETR)
		out_len = snprintf (out_buf, sizeof(out_buf),
				    "RETR %u\r\n", ++asked_for);
	    else if ((state == STAT && do_from) || state == TOP)
		out_len = snprintf (out_buf, sizeof(out_buf),
				    "TOP %u 0\r\n", ++asked_for);
	    else if(state == XDELE) {
		out_len = snprintf(out_buf, sizeof(out_buf),
				   "XDELE %u %u\r\n", 1, count);
		sent_xdele++;
	    }
	    else if(state == DELE)
		out_len = snprintf (out_buf, sizeof(out_buf),
				    "DELE %u\r\n", ++asked_deleted);
	    if (out_len < 0)
		errx (1, "snprintf failed");
	    if (push_write (s, ctx, out_buf, out_len) != out_len)
		err (1, "write");
	    if (verbose > 1)
		fprintf (stderr, "%s", out_buf);
	}
    }
    if (verbose)
	fprintf (stderr, "Done\n");
    if (do_from) {
	free (tmp);
	free (headers);
    } else {
	write_state_destroy (&write_state);
    }
    return 0;
}

#ifdef KRB5
/*
 * SASL AUTH in POP3 is quite simple, no USER or PASS is sent, rather
 * is the SASL authorization id mechanism used to transfer that
 * information.
 * 
 * Then the SASL authentication negotiation is done, integrity and/or
 * confidentiality is turned after the +OK message is received.  This
 * is so error can be returned from authentication.
 */

static int
do_sasl (const char *host,
	 int port,
	 const char *user,
	 const char *filename,
	 const char *header_str,
	 int leavep,
	 int verbose,
	 int forkp)
{
    char in_buf[PUSH_BUFSIZ + 1];	/* sentinel */
    size_t in_len = 0;
    char *in_ptr = in_buf;
    fd_set readset;
    int s, ret, flags;
    struct tsasl_peer *client;
    struct tsasl_prot *conn;
    int have_capa = 0;
    char *auth_sasl = NULL;

    enum { SASL_GREET, SASL_CAPA, SASL_SELECT_AUTH, 
	   SASL_AUTH, SASL_AUTH_OK, SASL_VERIFY_CAPA, 
	   SASL_DONE, SASL_FAILED } sasl_state;

    sasl_state = SASL_GREET;

    flags = TSASL_FLAGS_INITIATOR | 
	TSASL_FLAGS_BASE64_AUTH | 
	TSASL_FLAGS_INTEGRITY;

    if (do_encrypt)
	flags |= TSASL_FLAGS_CONFIDENTIALITY;

    if (verbose) {
	    fprintf(stderr, "Host: <%s>\n", host);
	    fprintf(stderr, "User: <%s>\n", user);
	    }

    if (tsasl_peer_init(flags, "pop", host, &client)) {
	warnx("peer init");
	return 1;
    }

    tsasl_authorization_id_set(client, user);

    s = do_connect (host, port, PF_UNSPEC, 1, &conn);
    if (s < 0) {
	tsasl_peer_free(client);
	return 1;
    }
    do {
	FD_ZERO(&readset);
#if 0
	if (s >= FD_SETSIZE)
	    errx (1, "fd too large");
#endif
	FD_SET(s,&readset);

	ret = select (s + 1, &readset, NULL, NULL, NULL);
	if (ret < 0) {
	    if (errno == EAGAIN)
		continue;
	    else
		err (1, "select");
	}

	if (FD_ISSET(s, &readset)) {
	    char *beg, *p;
	    size_t rem;

	    ret = tsasl_read (conn, 0, in_ptr, sizeof(in_buf) - in_len - 1);
	    if (ret == TSASL_PROT_EOF)
		errx (1, "EOF during read");
	    else if (ret < 0)
		errx (1, "tsasl_read");
	    
	    in_len += ret;
	    in_ptr += ret;
	    *in_ptr = '\0';
	    
	    beg = in_buf;
	    rem = in_len;
	    while(rem > 1
		  && (p = strstr(beg, "\r\n")) != NULL) {

		if (verbose > 1)
		    fprintf(stderr, "got: \"%.*s\"\n", (int)(p - beg), beg);

		if (sasl_state == SASL_AUTH) {
		    struct tsasl_buffer in, out;
		    char *p1;
		    
		    if (rem < 3 || beg[0] != '+') {
			warnx("got too short message in AUTH");
			sasl_state = SASL_FAILED;
		    }
		    p1 = beg + 1;
		    while (*p1 == ' ' && p1 < p)
			p1++;
		    in.tb_data = p1;
		    in.tb_size = p - p1;
		    if (in.tb_size < 0)
			in.tb_size = 0;
		    out.tb_size = 0;
		    
		    ret = tsasl_request(client, &in, &out);
		    
		    if (ret == TSASL_CONTINUE || out.tb_size != 0) {
			if (verbose > 1)
			    fprintf(stderr, "Send: %.*s\r\n",
				    (int)out.tb_size, (char *)out.tb_data);
			tsasl_printf(conn, TSASL_PROT_FLUSH, "%.*s\r\n",
				     (int)out.tb_size, (char *)out.tb_data);
			    }
		    
		    if (ret == TSASL_ERROR) {
			warnx("sasl error: %s\n", 
			      tsasl_get_errorstring(client));
			sasl_state = SASL_FAILED;
		    } else if (ret == TSASL_CONTINUE) {
			/* do nothing */;
		    } else if (ret == TSASL_DONE) {
			sasl_state = SASL_AUTH_OK;
		    } else {
			sasl_state = SASL_FAILED;
			warnx("unknown tsasl error %d\n", ret);
		    }
		} else if (rem > 3 && strncmp(beg, "+OK", 3) == 0) { 
		    if (sasl_state == SASL_GREET) {
			if (verbose > 1)
			    fprintf(stderr, "Send: CAPA\r\n");
			tsasl_printf(conn, TSASL_PROT_FLUSH, "CAPA\r\n");
			sasl_state++;
		    } else if (sasl_state == SASL_CAPA) {
			have_capa = 1;
		    } else if (sasl_state == SASL_VERIFY_CAPA) {
			if (!have_capa) {
			    warnx("didn't have CAPA before "
				  "auth but after: %.*s",
				  (int)(p - beg), beg);
			    sasl_state = SASL_FAILED;
			}
		    } else if (sasl_state == SASL_AUTH_OK) {
			tsasl_prot_peer(conn, client);
			sasl_state++;
			if (verbose > 1)
			    fprintf(stderr, "auth ok, trying CAPA again\n");
			if (verbose > 1)
			    fprintf(stderr, "CAPA\r\n");
			tsasl_printf(conn, TSASL_PROT_FLUSH, "CAPA\r\n");
		    } else
			warnx("unknown ok: %.*s\n", (int)(p - beg), beg);
		} else if (rem > 3 && strncmp(beg, "-ERR", 4) == 0) { 
		    if (sasl_state == SASL_CAPA) {
			sasl_state++;
		    } if (sasl_state == SASL_VERIFY_CAPA) {
			if (have_capa) {
			    warnx("had CAPA before auth but not after: %.*s",
				  (int)(p - beg), beg);
			    sasl_state = SASL_FAILED;
			}
			sasl_state++;
		    } else {
			warnx("%.*s\n", (int)(p - beg), beg); 
			sasl_state = SASL_FAILED;
		    }
		} else if (sasl_state == SASL_VERIFY_CAPA 
			   || sasl_state == SASL_CAPA)
		{
		    if (rem > 2 && strncmp(beg, ".\r\n", 3) == 0) {
			if (verbose > 1)
			    fprintf(stderr, "end of CAPA\n");
			sasl_state++;
		    } else if (rem > 4 && strncmp(beg, "SASL ", 5) == 0) {
			if (sasl_state == SASL_VERIFY_CAPA) {
			    if (auth_sasl == NULL)
				errx(1, "tried to verify "
				     "nonexisting CAPA SASL ?");
			    if (strncmp(auth_sasl, beg, (int)(p - beg)) != 0)
				errx(1, "CAPA SASL not same after AUTH");
			    if (verbose > 1)
				fprintf(stderr, "SASL CAPA same after AUTH\n");
			} else {
			    if (auth_sasl)
				warnx("got AUTH SASL twice");
			    auth_sasl = strndup(beg, (int)(p - beg));
			    if (auth_sasl == NULL)
				errx(1, "out of memory: %d", (int)(p - beg));
			}
		    } else if (verbose > 1)
			fprintf(stderr, "dropping unknonwn CAPA\n");
		} else
		    printf("ignoring: %.*s\n", (int)(p - beg), beg);

		in_len -= (p - beg) + 2;
		memmove(in_buf, p + 2, in_len);
		rem -= (p - beg) + 2;
		in_ptr -= (p - beg) + 2;

		if (sasl_state == SASL_SELECT_AUTH) {
		    char *mech = "GSSAPI";
		    if (auth_sasl) {
			/* XXX select best auth mech */
		    }
		    tsasl_select_mech(client, mech);
			    if (verbose > 1)
				    fprintf (stderr,  "Send: AUTH %s\r\n",
				  mech);
		    tsasl_printf (conn, TSASL_PROT_FLUSH, "AUTH %s\r\n",
				  mech);
		    sasl_state++;
		}
	    }
	}
    } while (sasl_state != SASL_FAILED && sasl_state != SASL_DONE);

    if (sasl_state == SASL_FAILED) {
	warnx("failed sasl");
	close(s);
	tsasl_peer_free(client);
	return 1;
    } else if (verbose) {
	char *msg;
	tsasl_status_get(client, &msg);
	printf("SASL ok: %s\n", msg);
	tsasl_status_free(msg);
    }

    ret = doit (s, conn, host, user, filename, header_str, leavep, 
		verbose, forkp,	PASS, in_buf, in_len);
	ret = 1;
    tsasl_prot_close(conn);
    close(s);
    tsasl_peer_free(client);
    return ret;
}

static int
do_v5 (const char *host,
       int port,
       const char *user,
       const char *filename,
       const char *header_str,
       int leavep,
       int verbose,
       int forkp)
{
    krb5_error_code ret;
    krb5_auth_context auth_context = NULL;
    krb5_principal server;
    struct tsasl_prot *conn;
    int s;

    s = do_connect (host, port, PF_UNSPEC, 1, &conn);
    if (s < 0)
	return 1;

    ret = krb5_sname_to_principal (context,
				   host,
				   "pop",
				   KRB5_NT_SRV_HST,
				   &server);
    if (ret) {
	warnx ("krb5_sname_to_principal: %s",
	       krb5_get_err_text (context, ret));
	return 1;
    }

    ret = krb5_sendauth (context,
			 &auth_context,
			 &s,
			 "KPOPV1.0",
			 NULL,
			 server,
			 0,
			 NULL,
			 NULL,
			 NULL,
			 NULL,
			 NULL,
			 NULL);
    krb5_free_principal (context, server);
    if (ret) {
	warnx ("krb5_sendauth: %s",
	       krb5_get_err_text (context, ret));
	return 1;
    }

    ret = doit (s, conn, host, user, filename, header_str, leavep, 
		verbose, forkp, INIT, "", 0);
	ret = 1;
    tsasl_prot_close(conn);
    close(s);
    return ret;
}
#endif

#ifdef KRB4
static int
do_v4 (const char *host,
       int port,
       const char *user,
       const char *filename,
       const char *header_str,
       int leavep,
       int verbose,
       int forkp)
{
    KTEXT_ST ticket;
    MSG_DAT msg_data;
    CREDENTIALS cred;
    des_key_schedule sched;
    int s;
    int ret;
    void *ctx;

    s = do_connect (host, port, PF_INET, 1, &ctx);
    if (s < 0)
	return 1;
    ret = krb_sendauth(0,
		       s,
		       &ticket, 
		       "pop",
		       (char *)host,
		       krb_realmofhost(host),
		       getpid(),
		       &msg_data,
		       &cred,
		       sched,
		       NULL,
		       NULL,
		       "KPOPV0.1");
    if(ret) {
	warnx("krb_sendauth: %s", krb_get_err_text(ret));
	return 1;
    }
    ret = doit (s, ctx, host, user, filename, header_str, 
		leavep, verbose, forkp, INIT, "", 0);
#if KRB5
    tsasl_prot_close(ctx);
#endif
    close(s);
    return ret;
}
#endif /* KRB4 */

#ifdef HESIOD

#ifdef HESIOD_INTERFACES

static char *
hesiod_get_pobox (const char **user)
{
    void *context;
    struct hesiod_postoffice *hpo;
    char *ret = NULL;

    if(hesiod_init (&context) != 0)
	err (1, "hesiod_init");

    hpo = hesiod_getmailhost (context, *user);
    if (hpo == NULL) {
	warn ("hesiod_getmailhost %s", *user);
    } else {
	if (strcasecmp(hpo->hesiod_po_type, "pop") != 0)
	    errx (1, "Unsupported po type %s", hpo->hesiod_po_type);

	ret = estrdup(hpo->hesiod_po_host);
	*user = estrdup(hpo->hesiod_po_name);
	hesiod_free_postoffice (context, hpo);
    }
    hesiod_end (context);
    return ret;
}

#else /* !HESIOD_INTERFACES */

static char *
hesiod_get_pobox (const char **user)
{
    char *ret = NULL;
    struct hes_postoffice *hpo;

    hpo = hes_getmailhost (*user);
    if (hpo == NULL) {
	warn ("hes_getmailhost %s", *user);
    } else {
	if (strcasecmp(hpo->po_type, "pop") != 0)
	    errx (1, "Unsupported po type %s", hpo->po_type);

	ret = estrdup(hpo->po_host);
	*user = estrdup(hpo->po_name);
    }
    return ret;
}

#endif /* HESIOD_INTERFACES */

#endif /* HESIOD */

static char *
get_pobox (const char **user)
{
    char *ret = NULL;

#ifdef HESIOD
    ret = hesiod_get_pobox (user);
#endif

    if (ret == NULL)
	ret = getenv("MAILHOST");
    if (ret == NULL)
	errx (1, "MAILHOST not set");
    return ret;
}

static void
parse_pobox (char *a0, const char **host, const char **user)
{
    const char *h, *u;
    char *p;
    int po = 0;

    if (a0 == NULL) {

	    fprintf(stderr, "No defaults\n");
	    exit(1);
#if 0
	*user = getenv ("USERNAME");
	if (*user == NULL) {
	    struct passwd *pwd = getpwuid (getuid ());

	    if (pwd == NULL)
		errx (1, "Who are you?");
	    *user = estrdup (pwd->pw_name);
	}
	*host = get_pobox (user);
	return;
#endif
    }

    /* if the specification starts with po:, remember this information */
    if(strncmp(a0, "po:", 3) == 0) {
	a0 += 3;
	po++;
    }
    /* if there is an `@', the hostname is after it, otherwise at the
       beginning of the string */
    p = strchr(a0, '@');
    if(p != NULL) {
	*p++ = '\0';
	h = p;
    } else {
	h = a0;
    }
    /* if there is a `:', the username comes before it, otherwise at
       the beginning of the string */
    p = strchr(a0, ':');
    if(p != NULL) {
	*p++ = '\0';
	u = p;
    } else {
	u = a0;
    }
    if(h == u) {
	/* some inconsistent compatibility with various mailers */
	if(po) {
	    h = get_pobox (&u);
	} else {
	    u = get_default_username ();
	    if (u == NULL)
		errx (1, "Who are you?");
	}
    }
    *host = h;
    *user = u;
}

static int
get_port(char *port_str, char *default_port_str, int default_port)
{
    int port = 0;

    if (port_str) {
	struct servent *s = roken_getservbyname (port_str, "tcp");
    
	if (s)
	    port = s->s_port;
	else {
	    char *ptr;
	    
	    port = strtol (port_str, &ptr, 10);
	    if (port == 0 && ptr == port_str)
		errx (1, "Bad port `%s'", port_str);
	    port = htons(port);
	}
    }

    if (port == 0) {
#ifdef KRB5
	port = krb5_getportbyname (context, default_port_str, 
				   "tcp", default_port);
#elif defined(KRB4)
	port = k_getportbyname (default_port_str, "tcp", htons(default_port));
#else
#error must define KRB4 or KRB5
#endif
    }
    return port;
}

int
main(int argc, char **argv)
{
    int sport = 0;
    int kport = 0;
    int optind = 0;
    int ret = 1;
    const char *host, *user, *filename = NULL;
    char *pobox = NULL;

    setprogname (argv[0]);

#ifdef KRB5
    {
	krb5_error_code ret;

	ret = krb5_init_context (&context);
	if (ret)
	    errx (1, "krb5_init_context failed: %d", ret);
    }
#endif

    if (getarg (args, sizeof(args) / sizeof(args[0]), argc, argv,
		&optind))
	usage (1);

    argc -= optind;
    argv += optind;

#if defined(KRB4) && defined(KRB5)
    if(use_v4 == -1 && use_v5 == 1)
	use_v4 = 0;
    if(use_v5 == -1 && use_v4 == 1)
	use_v5 = 0;
#endif    

    if (do_help)
	usage (0);

#if 0
    if (do_version) {
	print_version(NULL);
	return 0;
    }
#endif
	
    if (do_from && header_str == NULL)
	header_str = "From:";
    else if (header_str != NULL)
	do_from = 1;

    if (do_from) {
	if (argc == 0)
	    pobox = NULL;
	else if (argc == 1)
	    pobox = argv[0];
	else
	    usage (1);
    } else {
	if (argc == 1) {
	    filename = argv[0];
	    pobox    = NULL;
	} else if (argc == 2) {
	    filename = argv[1];
	    pobox    = argv[0];
	} else
	    usage (1);
    }

    kport = get_port(kport_str, "kpop", 1109);
    sport = get_port(sport_str, "pop3", 110);

    parse_pobox (pobox, &host, &user);

#ifdef KRB5
    if (ret && use_sasl) {
	ret = do_sasl (host, sport, user, filename, header_str,
		       do_leave, verbose_level, do_fork);
    }

    if (ret && use_v5) {
	ret = do_v5 (host, kport, user, filename, header_str,
		     do_leave, verbose_level, do_fork);
    }
#endif

#ifdef KRB4
    if (ret && use_v4) {
	ret = do_v4 (host, kport, user, filename, header_str,
		     do_leave, verbose_level, do_fork);
    }
#endif /* KRB4 */
    return ret;
}
