/*
 * Copyright (c) 2004 Thomas Nystrm and Stacken Computer Club
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "common.h"
#pragma hdrstop

#include <tsasl.h>
#include "tsaslw.h"
#include "values.h"

void LogText(const char *fmt, ...);

TSasl::TSasl(const char *_host, const char *_user)
{
	Host = _host;
	User = _user;
	sasl_state = SASL_GREET;
	flags = TSASL_FLAGS_INITIATOR | 
		TSASL_FLAGS_BASE64_AUTH | 
		TSASL_FLAGS_INTEGRITY;

	flags |= TSASL_FLAGS_CONFIDENTIALITY;
	client = NULL;
	ExitSemaphore = NULL;
	EncryptOn = false;
	
}

TSasl::~TSasl()
{
	if (client != NULL)
		tsasl_peer_free(client);
}

void
TSasl::ProtocolInit(SocketChannel *_sock, TEventSemaphore *_ExitSemaphore)
{
	int ret;

	sock = _sock;
	ExitSemaphore = _ExitSemaphore;
	rcv_valid = rcv_pek = 0;

	if (tsasl_peer_init(flags, "pop", Host, &client))
		throw TSaslError("peer init");

	tsasl_authorization_id_set(client, User);

	ret = tsasl_prot_init(sock->sock, &ctx);
	if (ret)
		throw TSaslError("tsasl_prot_init: failed with", ret);
}

char *
TSasl::Authenticate()
{
	char *line = NULL, *beg;
	int have_capa = 0;
	char *auth_sasl = NULL;
	int ret;
	char *greeting = NULL;

	try {
		while (sasl_state != SASL_DONE) {

			GetLine(&line);

			LogText("<<< %s\n", line);

			beg = line;
			if (sasl_state == SASL_AUTH) {
				struct tsasl_buffer in, out;
				char *p1;

				if (*beg != '+')
					throw TSaslError("Bad response to AUTH");

				p1 = beg + 1;
				while (*p1 == ' ')
					p1++;
				in.tb_data = p1;
				in.tb_size = strlen(p1);
				if (in.tb_size < 0)
					in.tb_size = 0;
				out.tb_size = 0;

				ret = tsasl_request(client, &in, &out);

				if ((ret == TSASL_CONTINUE) || (out.tb_size != 0))
					SendCmd("%.*s", (int)out.tb_size, (char *)out.tb_data);

				if (ret == TSASL_ERROR)
					throw TSaslError("sasl error", 
							 tsasl_get_errorstring(client));
				else if (ret == TSASL_CONTINUE) {
						/* do nothing */;
					}
				else if (ret == TSASL_DONE) {
					sasl_state = SASL_AUTH_OK;
					}
				else
					throw TSaslError("unknown tsasl error",
							 ret);

				}
			else if (strncmp(beg, "+OK", 3) == 0) { 
				if (sasl_state == SASL_GREET) {
					greeting = strdup(beg);
					SendCmd("CAPA");
					sasl_state++;
					}
				else if (sasl_state == SASL_CAPA) {
					have_capa = 1;
					}
				else if (sasl_state == SASL_VERIFY_CAPA) {
					if (!have_capa)
						throw TSaslError("didn't have CAPA before auth but after",
								 beg);
					}
				else if (sasl_state == SASL_AUTH_OK) {
					tsasl_prot_peer(ctx, client);
					EncryptOn = true;
					sasl_state++;
					LogText("auth ok, trying CAPA again\n");
					SendCmd("CAPA");
					}
				else
					LogText("unknown ok: %*s\n", beg);
				}
			else if (strncmp(beg, "-ERR", 4) == 0) { 
				if (sasl_state == SASL_CAPA) {
					sasl_state++;
					}
				if (sasl_state == SASL_VERIFY_CAPA) {
					if (have_capa)
						throw TSaslError("had CAPA before auth but not after",
								 beg);
					sasl_state++;
					}
				else
					throw TSaslError(beg);
				}
			else if ((sasl_state == SASL_VERIFY_CAPA) ||
				 (sasl_state == SASL_CAPA)) {
				if (strcmp(beg, ".") == 0) {
					LogText("end of CAPA\n");
					sasl_state++;
					}
				else if (strncmp(beg, "SASL ", 5) == 0) {
					if (sasl_state == SASL_VERIFY_CAPA) {
						if (auth_sasl == NULL)
							throw TSaslError("tried to verify nonexisting CAPA SASL?");
						if (strcmp(auth_sasl, beg) != 0)
							throw TSaslError("CAPA SASL not same after AUTH");
						LogText("SASL CAPA same after AUTH\n");
						}
					else {
						if (auth_sasl)
							LogText("got AUTH SASL twice");
						auth_sasl = strdup(beg);
						if (auth_sasl == NULL)
							throw TSaslError("Out of memory");
						}
					}
				else
					LogText("dropping unknonwn CAPA\n");
				}
			else
				LogText("Ignorerar: %s\n", beg);

			if (sasl_state == SASL_SELECT_AUTH) {
				char *mech = "GSSAPI";
				if (auth_sasl) {
					/* XXX select best auth mech */
					}
				tsasl_select_mech(client, mech);
				SendCmd("AUTH %s", mech);
				sasl_state++;
				}

			free(line);
			line = NULL;

			}
		}
	catch (...) {
		if (line)
			free(line);
		if (greeting)
			free(greeting);
		throw;
		}

	return greeting;
}

void
TSasl::GetLine(char **line)
{
	int maxchar;
	register int p = 0;
	int ch;

	maxchar = 128;
	if ((*line = (char *)malloc(maxchar+1)) == NULL)
		throw TSaslError("Out of memory");

	while ((ch = Getc()) != EOF) {
		switch (ch) {
			case -2:
			case 0:
				break;	// Ignore!
			case 0x0d:	// Enter	/* Ignore Enter */
				break;
			case 0x0a:	// NewLine
				(*line)[p] = 0;
				return;
			default:
				if (p >= maxchar) {
					maxchar += 128;
					if ((*line = (char *)realloc(*line, maxchar+1)) == NULL)
						throw TSaslError("Out of memory");
					}
				if (ch >= ' ')
					(*line)[p++] = (char)ch;

			}
		}
	if (ch == EOF)
		throw TSaslError("Closed");

	return;
}

bool
TSasl::WaitForEvent(SocketChannel &sock, ulong timeOut)
{
	TSemaphoreSet ss(0, 2);
	
	ss.Add(*ExitSemaphore);
	ss.Add(*sock.event);
	
	TSemaphoreSet::TLock l(ss, TSemaphoreSet::WaitAny, timeOut);
	if ((l.WhichAquired() == TSemaphoreSet::TLock::TimedOut) &&
	    (timeOut != INFINITE))
		return true;

	if (!l.WasAquired())
		return false;		// Something failed!

	if (l.WhichAquired() != 1)
		return false;		// Didn't get expected event
			
	return true;
}

bool
TSasl::FillBuffer(bool WaitFlag)
{
	int this_transfer;

      one_more_time:
	if (WaitFlag)
		if (!WaitForEvent(*sock))
			return false;

	if ((this_transfer = ::recv(sock->sock, rcv_buf, sizeof(rcv_buf), 0)) <= 0) {
		if ((this_transfer == SOCKET_ERROR) &&
		    (WSAGetLastError() == WSAEWOULDBLOCK)) {
			if (WaitFlag) {
				sock->event->Reset();
				goto one_more_time;
				}
			else {
				rcv_pek = rcv_valid = 0;
				return true;
				}
			}
		if (this_transfer == SOCKET_ERROR)
			throw TSaslError("Socketerror", WSAGetLastError());
		closesocket(sock->sock);
		sock->sock = INVALID_SOCKET;
		return false;
		}

	if (EncryptOn) {
		int ret;
		struct tsasl_buffer in, out;

		out.tb_data = strdup("");
		out.tb_size = 0;

		in.tb_data = rcv_buf;
		in.tb_size = this_transfer;

		ret = tsasl_unwrap(client, &in, &out);
		if ((ret == 0) && (out.tb_size != 0)) {
			memcpy(rcv_buf, out.tb_data, out.tb_size);
			rcv_valid = out.tb_size;
			tsasl_buffer_free(&out);
			}
		else
			rcv_valid = 0;
		}
	else
		rcv_valid = this_transfer;

	rcv_pek = 0;
	return true;
}

int
TSasl::Getc()
{
	if (rcv_pek >= rcv_valid) {
		if (!FillBuffer(true))
			return EOF;
		}

	if (rcv_pek >= rcv_valid)
		return -2;

	return (rcv_buf[rcv_pek++] & 0xff);
}

size_t
TSasl::Read(unsigned char *buf, size_t max)
{
	int size = 0;

	if (rcv_pek >= rcv_valid)
		if (!FillBuffer(false))
			return -1;
	
	if (rcv_pek < rcv_valid) {
		size = rcv_valid - rcv_pek;
		if (size > max)
			size = max;
		memcpy(buf, &rcv_buf[rcv_pek], size);
		rcv_pek += size;
		}
	return size;
}

size_t
TSasl::Write(const unsigned char *buf, size_t size)
{
	return tsasl_write(ctx, TSASL_PROT_FLUSH, (char *)buf, size);
}

void
TSasl::SendCmd(const char *fmt, ...)
{
	va_list va;
	char *msg;
	ssize_t ret;

	va_start(va, fmt);
	ret = vasprintf(&msg, fmt, va);
	va_end(va);

	if (ret < 0)
		return;

	LogText(">>> %s\n", msg);

	tsasl_write(ctx, 0, msg, strlen(msg));
	tsasl_write(ctx, TSASL_PROT_FLUSH, "\r\n", 2);

	free(msg);

}

//****************************************************************

TSaslError::TSaslError(const char *s)
	: xmsg(MakeString(s, NULL))
{}

TSaslError::TSaslError(const char *s, int errorCode)
	: xmsg(MakeString(s, itoa(errorCode, pbuf, 10)))
{}

TSaslError::TSaslError(const char *s, const char *s1)
	: xmsg(MakeString(s, s1))
{}

string
TSaslError::MakeString(const char *s, const char *s1)
{
	string Msg;

	Msg.reserve(strlen(s) + 32);

	Msg = "";
	if (*s != '-')
		Msg += "-ERR ";

	Msg += s;

	if (s1) {
		Msg += ": ";
		Msg += s1;
		}
	LogText("TSaslError: %s\n", Msg.c_str());
	return Msg;
}

