/*
 * Copyright (c) 1999-2002 Thomas Nystrm and Stacken Computer Club
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by the Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

/* 
 * Based on work by: Jrgen Karlsson - d93-jka@nada.kth.se
 */

#include "common.h"
#pragma hdrstop

#include "KClient.h"
#include "kclient.rh"

#include <krb.h>
#include <setup.rh>

#include "passdia.h"

//****************************************************************

//****************************************************************

class CrossRealm4 : public CrossRealm
{
      public:
	CrossRealm4(KClient *parent, const char *user, const char *inst,
		char *realm, size_t realm_siz, int crossmode,
		SetupParam *setupData);

	virtual bool LoggedIn(const char *realm);
	virtual int Login(KCPBuf &b, const char *alt_realm = 0);
	virtual int GetAltRealm(const char *realm, char *alt_realm);
	virtual int TryCrossAuth(const char *_realm);
	virtual char **GetRealms();
	const char *ErrorText(int err, char *text);

	KClient *KParent;
};

//****************************************************************

class TicketFile
{
      public:
	TicketFile();
	~TicketFile();

	bool Open();
	void Close();
	bool Read(CREDENTIALS &c, const char *service = 0,
		const char *instance = 0, const char *realm = 0);
	bool Opened;
};

//****************************************************************

Kerr
KClient::Message(const char *text, Kerr error)
{
	HWND h = 0;

	if (Parent)
		h = *Parent;

	::MessageBox(h, text, errorHeader, MB_OK|MB_ICONEXCLAMATION);

	return error;
}

/* KClient::KClient
 * The local and remote addresses are for use in KClientEncrypt,
 * KClientDecrypt, KClientMakeSendAuth and KClientVerifySendAuth.
 * If you don't use any of these routines it's perfectly OK to ignore them
 */
KClient::KClient(TWindow *parent, SetupParam *setupData,
	unsigned long _lAddr, unsigned short _lPort,
	unsigned long _fAddr, unsigned short _fPort)
{
	Parent = parent;
	SetupData = setupData;

	SlAddr = _lAddr;
	SlPort = _lPort;
	SfAddr = _fAddr;
	SfPort = _fPort;
	if (tf_get_pname(Suser) != KSUCCESS)
		Suser[0] = 0;

	if (tf_get_pinst(Sinst) != KSUCCESS)
		Sinst[0] = 0;
	krb_get_lrealm(Srealm, 1);
}

KClient::~KClient()
{
}

/* SetDefaultPrincipal
 * Parse 'princ' as a principal: user[.instance][@realm]
 * and store it for later use as defaults in password dialog box
 */
Kerr
KClient::SetDefaultPrincipal(const char *princ)
{
	char user[ANAME_SZ] = "";
	char instance[INST_SZ] = "";
	char realm[REALM_SZ] = "";
	int err;

	if (princ) {
		if (princ[0])
			if ((err = kname_parse(user, instance, realm, (char *)princ)) != 0)
				return Message(ErrorText(err,0),err);
	
		strlcpy(Suser, user, ANAME_SZ);
		strlcpy(Sinst, instance, INST_SZ);
		strlcpy(Srealm, realm, REALM_SZ);
		return 0;
		}

	if (tf_get_pname(Suser) != KSUCCESS)
		Suser[0] = 0;
	if (tf_get_pinst(Sinst) != KSUCCESS)
		Sinst[0] = 0;
	krb_get_lrealm(Srealm, 1);
	return 0;
}

/* KClientGetTicketForService
 * This routine gets an authenticator to be passed to a service.
 * If the user isn't already logged in the user is prompted for a password.
 */
Kerr
KClient::GetTicketForService(char *service, void *buf, unsigned long *buflen,
	int checksum, bool NoMsg)
{
	CREDENTIALS c;
	KClientKey k;
	KTEXT_ST ticket;
	char serv[255], inst[255], realm[255];
	Kerr err;

	// KClientSetUserName(Suser);
	kname_parse(serv,inst,realm,service);
	if (*realm)
		strcpy(Srealm,realm);
	else
		strcpy(realm,Srealm);
	if (Status() == KClientNotLoggedIn)
		if ((err = Login(&k, NoMsg)) != KSUCCESS)
			return err;

	if ((err = krb_mk_req(&ticket, serv, inst, realm, checksum)) != KSUCCESS) {
		if (NoMsg)
			return err;
		return Message(ErrorText(err,0),err);
		}
	if ((err = krb_get_cred(serv, inst, realm, &c)) != KSUCCESS) {
		if (NoMsg)
			return err;
		return Message(ErrorText(err,0),err);
		}

	if (*buflen >= (ticket.length + sizeof(unsigned long))) {
		*buflen = ticket.length + sizeof(unsigned long);
		CopyMemory(buf, &ticket, *buflen);
		CopyMemory(Skey, c.session, sizeof(Skey));
		strlcpy(Suser, c.pname, ANAME_SZ);
		strlcpy(Sinst, c.pinst, INST_SZ);
		strlcpy(Srealm, c.realm, REALM_SZ);
		}
	else
		err = -1;

	return err;
}


/* KClientLogin
 * This routine "logs in" by getting a ticket granting ticket from kerberos.
 * It returns the user's private key which can be used to automate login at
 * a later time with KClientKeyLogin.
 */

Kerr
KClient::Login(KClientKey *privateKey, bool NoMsg)
{
	CREDENTIALS c;
	Kerr err;

	int CrossMode = GetResourceIdxTable(SETUP_SIT_CROSSMODE,
		SetupData->InterrealmMode);

	if ((err = CrossRealm4(this, Suser, Sinst, Srealm, sizeof(Srealm),
		CrossMode, SetupData).Execute()) != 0) {
		if (NoMsg)
			return err;
		return Message(ErrorText(err,0),err);
		}

	if ((err = krb_get_cred(KRB_TICKET_GRANTING_TICKET, Srealm,
				0, &c)) == KSUCCESS)
		CopyMemory(privateKey, c.session, sizeof(*privateKey));


	if (err == KSUCCESS) {
		strlcpy(Suser, c.pname, sizeof(Suser));
		strlcpy(Sinst, c.pinst, sizeof(Sinst));
		}

	if (NoMsg || (err == KSUCCESS))
		return err;

	return Message(ErrorText(err,0),err);
}


/* KClientPasswordLogin
 * This routine is similiar to KClientLogin but instead of prompting the user
 * for a password it uses the password supplied to establish login.
 */
#pragma argsused
Kerr
KClient::PasswordLogin(char *password, KClientKey *privateKey)
{
	return krb_get_pw_in_tkt(Suser, Sinst, Srealm,
				 KRB_TICKET_GRANTING_TICKET, Srealm, DEFAULT_TKT_LIFE,
				 password);
}

#pragma argsused
static int
_stdcall
key_proc(const char *arg, char *a2, const char *a3, const void *a4, des_cblock *a5)
{
	return 0;
}

/* KClientKeyLogin
 * This routine is similiar to KClientLogin but instead of prompting the user
 * for a password it uses the private key supplied to establish login.
 */
Kerr
KClient::KeyLogin(KClientKey *privateKey)
{
    return krb_get_in_tkt(Suser, Sinst, Srealm,
			  KRB_TICKET_GRANTING_TICKET,
			  Srealm,
			  DEFAULT_TKT_LIFE,
			  key_proc,
			  0,
			  privateKey);
}

/* KClientLogout
 * This routine destroys all credentials stored in the credential cache
 * effectively logging the user out.
 */
Kerr
KClient::Logout(void)
{
	return 0;
}


/* KClientStatus
 * This routine returns the user's login status which can be
 * KClientLoggedIn or KClientNotLoggedIn.
 */
short
KClient::Status(const char *realm)
{
	CREDENTIALS c;

	if (!realm)
		realm = Srealm;

	TicketFile tf;
	tf.Open();
	return (short)(tf.Read(c, KRB_TICKET_GRANTING_TICKET, realm)
		? KClientLoggedIn
		: KClientNotLoggedIn);
}


/* KClientGetUserName
 * This routine returns the name the user supplied in the login dialog.
 * No name is returned if the user is not logged in.
 */
Kerr
KClient::GetUserName(char *user)
{
	strcpy(user, Suser);
	return 0;
}


/* KClientSetUserName
 * This routine sets the name that will come up in the login dialog
 * the next time the user is prompted for a password.
 */
Kerr
KClient::SetUserName(const char *user)
{
	strcpy(Suser, user);

	return 0;
}


/* KClientCacheInitialTicket
 * This routine is used to obtain a ticket for the password changing service.
 */
#pragma argsused
Kerr
KClient::CacheInitialTicket(char *service)
{
	return 0;
}


/* KClientGetSessionKey
 * This routine can be used to obtain the session key which is stored
 * in the KClientSessionInfo record. The session key has no usefullness
 * with any KClient calls but it can be used to with the MIT kerberos API.
 */
Kerr
KClient::GetSessionKey(KClientKey *sessionKey)
{
	CopyMemory(sessionKey, Skey, sizeof(*sessionKey));
	return 0;
}


/* KClientMakeSendAuth
 * This routine is used to create an authenticator that is the same as those
 * created by the kerberos routine SendAuth.
 */
#pragma argsused
Kerr
KClient::MakeSendAuth(char *service,
		    void *buf,
		    unsigned long *buflen,
		    long checksum,
		    char *applicationVersion)
{
	return 0;
}


/* KClientVerifySendAuth
 * This routine is used to verify a response made by a server doing RecvAuth.
 * The two routines KClientMakeSendAuth and KClientVerifySendAuth together
 * provide the functionality of SendAuth minus the transmission of
 * authenticators between client->server->client.
 */
#pragma argsused
Kerr
KClient::VerifySendAuth(void *buf, unsigned long *buflen)
{
	return 0;
}


/* KClientEncrypt
 * This routine encrypts a series a bytes for transmission to the remote host.
 * For this to work properly you must be logged in and you must have specified
 * the remote and local addresses in KClientInitSession. The unencrypted
 * message pointed to by buf and of length buflen is returned encrypted
 * in encryptBuf of length encryptLength.
 * The encrypted buffer must be at least 26 bytes longer the buf.
 */
#pragma argsused
Kerr
KClient::Encrypt(unsigned char *buf, unsigned long buflen,
	unsigned char *encryptBuf, unsigned long *encryptLength)
{
	int num = 64;
	des_cfb64_encrypt(buf, encryptBuf, buflen,
		      (struct des_ks_struct*) Skey,
		      0, &num, 1);
	return 0;
}


/* KClientDecrypt
 * This routine decrypts a series of bytes received from the remote host.

 * NOTE: this routine will not reverse a KClientEncrypt call.
 * It can only decrypt messages sent from the remote host.

 * Instead of copying the decrypted message to an out buffer,
 * the message is decrypted in place and you are returned
 * an offset into the buffer where the decrypted message begins.
 */
Kerr
KClient::Decrypt(unsigned char *buf, unsigned long buflen,
	unsigned long *decryptOffset, unsigned long *decryptLength)
{
	int num;
	des_cfb64_encrypt(buf, buf, buflen,
		      (struct des_ks_struct*)Skey, 0, &num, 0);
	*decryptOffset = 0;
	*decryptLength = buflen;
	return 0;
}


/* KClientErrorText
 * This routine returns a text description of errors returned by any of
 * the calls in this library.
 */
const char *
KClient::ErrorText(Kerr err, char *text)
{
	const char *t = krb_get_err_text(err);

	if (text)
		strcpy(text, t);

	return t;
}

//****************************************************************

CrossRealm4::CrossRealm4(KClient *parent, const char *user,
	const char *inst, char *realm, size_t realm_siz,
	int crossmode, SetupParam *setupData)
	: CrossRealm(parent->Parent, user, inst, realm, realm_siz, crossmode,
		setupData)
{
	KParent = parent;
}

bool
CrossRealm4::LoggedIn(const char *realm)
{
	CREDENTIALS c;

	if (!realm)
		realm = Realm;

	TicketFile tf;
	tf.Open();
	return tf.Read(c, KRB_TICKET_GRANTING_TICKET, realm);
}

Kerr
CrossRealm4::Login(KCPBuf &b, const char *alt_realm)
{
	wkrb_dest_tkt_realm(b.realm);

	return krb_get_pw_in_tkt(b.user, b.instance, b.realm,
		KRB_TICKET_GRANTING_TICKET, alt_realm ? alt_realm : b.realm,
		DEFAULT_TKT_LIFE, b.passwd);
}

Kerr
CrossRealm4::GetAltRealm(const char *realm, char *alt_realm)
{
	return wkrb_get_alt_realm(realm, alt_realm);
}

Kerr
CrossRealm4::TryCrossAuth(const char *_realm)
{
	char lrealm[REALM_SZ];
	Kerr err;
	CREDENTIALS c;

	strlcpy(lrealm, _realm, REALM_SZ);

	err = krb_get_cred_kdc(KRB_TICKET_GRANTING_TICKET,
		Realm, lrealm, 255, &c);

	if (err == KSUCCESS) {
		err = save_credentials_cred(&c);
		memset(&c, 0, sizeof(c));
		if (err == KSUCCESS)
			err = krb_get_cred(KRB_TICKET_GRANTING_TICKET,
				Realm, lrealm, &c);
		}

	return err;
}

typedef char *charp;

char **
CrossRealm4::GetRealms()
{
	register int i, j;
	TicketFile tf;
	CREDENTIALS c;
	char **RealmTab;
	
	RealmTab = NULL;
	tf.Open();
	i = 0;
	while (tf.Read(c, KRB_TICKET_GRANTING_TICKET)) {
		if (strcmp(c.instance, c.realm) == 0)
			i++;
		}
	if (i) {
		RealmTab = new charp[i+1];
		ZeroMemory(RealmTab, (i + 1) * sizeof(char *));
		
		tf.Close();
		tf.Open();

		j = 0;

		while (tf.Read(c, KRB_TICKET_GRANTING_TICKET) && (j < i)) {
			if (strcmp(c.instance, c.realm) == 0) {
				RealmTab[j++] = strnewdup(c.realm);
				}
			}
		RealmTab[j] = 0;
		}
	return RealmTab;
}

const char *
CrossRealm4::ErrorText(Kerr err, char *text)
{
	return KParent->ErrorText(err, text);
}

//****************************************************************

TicketFile::TicketFile()
{
	Opened = false;
}

TicketFile::~TicketFile()
{
	Close();
}

bool
TicketFile::Open()
{
	if (!Opened)
		Opened = (tf_init(TKT_FILE, R_TKT_FIL) == KSUCCESS);

	return Opened;
}

void
TicketFile::Close()
{
	if (Opened) {
		tf_close();
		Opened = false;
		}
}

bool
TicketFile::Read(CREDENTIALS &c, const char *service,
	const char *instance, const char *realm)
{
	if (!Opened)
		return false;

	while (tf_get_cred(&c) == KSUCCESS) {
		if ((service ? (strcmp(c.service,service) == 0) : true) &&
		    (instance ? (strcmp(c.instance,instance) == 0) : true) &&
		    (realm ? (strcmp(c.realm,realm) == 0) : true))
		return true;
		}
	
	return false;
}

