/*
 * Copyright (c) 1999-2001 Thomas Nystrm and Stacken Computer Club
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "common.h"
#pragma hdrstop

#include "ftpmach.h"
#include "ftpconn.h"
#include "emul.rh"
#include "ftp.rh"

extern Kerberos5Context *k5c;

//****************************************************************

FTPConnection::FTPConnection(TWindow *parent)
	: Connection(parent, &::SetupData, &::DebugData)
{
	Principal = 0;

	IconId = IDI_FTP;

	fm = new FTPMachine(this, parent->GetApplication());
	cb = new FTPCmdBuffer(this, fm);
	ControlStream = 0;
	fm->Start();
}

FTPConnection::~FTPConnection()
{
	TerminalNote("\r\n[ FTP Session closed ]\r\n");

	if (Principal)
		delete[] Principal;

	switch (fm->GetStatus()) {
		case TThread::Created:
			fm->Start();
		case TThread::Running:
			fm->TerminateAndWait();
			break;
		case TThread::Suspended:
			fm->Resume();
			fm->TerminateAndWait();
			break;
		}

	if (ControlStream)
		delete ControlStream;

	delete fm;
	delete cb;

}

#pragma argsused
bool
FTPConnection::Retry(const char *errmsg, const char *realm)
{
	char buf[1024];
	char *bp = buf;

	if (errmsg) {
		strcpy(buf, errmsg);
		strcat(buf, "\n");
		bp = &buf[strlen(buf)];
		}

	if (realm)
		snprintf(bp, sizeof(buf) - strlen(bp),
			ResString(Parent, IDS_REOPEN), realm);
	else
		snprintf(bp, sizeof(buf) - strlen(bp),
			ResString(Parent, IDS_REOPEN2));

	if (!SetupData->DisableWarningBox) {
		if (Parent->MessageBox(buf,
			header, MB_YESNO|MB_ICONQUESTION) != IDYES) {
			fm->RetryOp = FTPMachine::ftpRetry;
			return false;
			}
		}
	ControlStream->GetSecurity()->DeleteTickets();
	fm->RetryOp = FTPMachine::ftpAbort;
	return true;
}

Connection *
FTPConnection::Reopen()
{
	Connection *nc;
	char buf[1024];

	fm->RetryOp = FTPMachine::ftpAbort;

	nc = new FTPConnection(Parent);
	if (nc->Open(cd)) {
		strlcpy(buf, header, sizeof(buf));
		strlcat(buf, " - ", sizeof(buf));
		strlcat(buf, nc->GetConnectionName(), sizeof(buf));
		Parent->Parent->SetCaption(buf);
		if (Principal)
			nc->SetDefaultPrincipal(Principal);
		return nc;
		}

	delete nc;
	return 0;
}

bool
FTPConnection::Open(ConnectionData _cd)
{
	bool res;

	Connection::Open(_cd);

	if (ControlStream) {
		fm->AssignControlStream(NULL);
		delete ControlStream;
		}

	ControlStream = new FTPControlStream(Parent, fm, SetupData,
		DebugData);
	res = ControlStream->Open(cd);
	fm->AssignControlStream(ControlStream);

	strlcpy(cd.HostName, ControlStream->GetHostName(),
		sizeof(cd.HostName));

	return res;
}

bool
FTPConnection::IsOpen()
{
	if (ControlStream)
		return ControlStream->IsOpen();

	return false;
}

bool
FTPConnection::IsLocal()
{
	return false;
}

void
FTPConnection::SetDefaultPrincipal(const char *princ)
{
	if (ControlStream) {
		ControlStream->SetDefaultPrincipal(princ);
		if (Principal)
			delete[] Principal;
		Principal = strnewdup(princ);
		}
}

const char *
FTPConnection::GetConnectionName()
{
	return cd.HostName;
}

ConnectionType 
FTPConnection::GetConnectionType()
{
	return ctFTP;
}

#pragma argsused
void
FTPConnection::InvokeCommand(unsigned short command, void* Data)
{
}

void
FTPConnection::Send(const char *str, unsigned int len, bool)
{
	cb->PutChars((const uchar *)str, len);
}

bool
FTPConnection::SendPossible()
{
	return true;
}


void
FTPConnection::SendRetryConnection(const char *msg)
{
	Connection::SendRetryConnection(msg);
}

void
FTPConnection::CloseConnection()
{
	Parent->PostMessage(::RegisterWindowMessage(WM_REG_CONNREAD), 0, 0);

	fm->AssignControlStream(NULL);
	if (ControlStream) {
		delete ControlStream;
		ControlStream = 0;
		}

//	Parent->PostMessage(WM_COMMAND, CM_CLOSECONN, 0);
}

bool
FTPConnection::ProcessReceived(WPARAM wParam)
{
	if (wParam) {
		if (ControlStream)
			ControlStream->ProcessReceived(wParam);
		return false;
		}
	return NewCharsReceived();
}
