/*
 * Copyright (c) 1998-2000 Thomas Nystrm and Stacken Computer Club
 * (Royal Institute of Technology, Stockholm, Sweden).
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by Kungliga Tekniska
 *      Hgskolan and its contributors.
 * 
 * 4. Neither the name of the Institute nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include "common.h"
#pragma hdrstop

#include <owl\inputdia.h>
#include "tickets.h"
#include "tickets.rh"

#include "kclient.h"
#include <krb.h>
#include <kerb5.h>

#include <kadm.h>
#include <kadm_err.rh>

extern Kerberos5Context *k5c;

static char pname[ANAME_SZ], pinst[INST_SZ], prealm[REALM_SZ];
static char NewTitle[128];

static char *
printable_time(time_t t)
{
	static char s[128];
	struct tm *tm;

	if ((tm = localtime(&t)) == NULL)
		strlcpy(s, "?????????", sizeof(s));
	else
		strftime(s, sizeof(s), "%c", tm);
	return s;
}

//****************************************************************

#define CPW_MAXSTR 64

class ChangePassword : public TDialog
{
      public:
	ChangePassword(TWindow *parent, int captionId = 0);
	~ChangePassword();

	virtual void SetupWindow();
	virtual void CloseWindow(int code);

      protected:
	virtual const char *CheckPassword(char *pword) = 0;
	virtual char *DoPSWChange() = 0;
	virtual void ErrorS(int ResId, const char *kerr = 0,
		const char *note = 0);

	struct {
		char user[CPW_MAXSTR];
		char inst[INST_SZ];
		char realm[REALM_SZ];
		char old[CPW_MAXSTR];
		char pnew[CPW_MAXSTR];
		char ver[CPW_MAXSTR];
	} trfbuf;
	int CaptionId;
	char SpecTktFile[128];
	HCURSOR OldCur;
};
//****************************************************************
class ChangePasswordK4 : public ChangePassword
{
      public:
	ChangePasswordK4(TWindow *parent);
	~ChangePasswordK4();

      protected:
	virtual const char *CheckPassword(char *pword);
	virtual char *DoPSWChange();
	virtual void Error(int ResId, int kerr = 0, const char *note = 0);

	void *SavedTicketFile;
};

//****************************************************************
class ChangePasswordK5 : public ChangePassword
{
      public:
	ChangePasswordK5(TWindow *parent);
	~ChangePasswordK5();

      protected:
	virtual const char *CheckPassword(char *pword);
	virtual char *DoPSWChange();
	virtual void Error(int ResId, int kerr = 0, const char *note = 0);

	Kerberos5Context *k5c;
};

//****************************************************************

DEFINE_RESPONSE_TABLE1(TicketsWindow, WXFrameWindow)
	EV_COMMAND(CM_KRBREFRESH, CmRefresh),
	EV_COMMAND(CM_KRBCLEAR, CmClear),		
	EV_COMMAND(CM_KRBLOGIN, CmLogin),
	EV_COMMAND(CM_KRB5LOGIN, CmLogin5),
	EV_COMMAND(CM_HIDE, EvClose),
	EV_COMMAND(CM_EXIT, CmExit),
	EV_COMMAND(CM_KRBCPW4, CmCPW4),
	EV_COMMAND(CM_KRBCPW5, CmCPW5),
	EV_COMMAND(CM_TEST, CmTest),
	EV_REGISTERED(WM_REG_TICKETUPDATE, EvUpdateCache),
	EV_REGISTERED(WM_REG_TRAYNOTE, TrayNotify),
	EV_WM_ACTIVATE,		
	EV_WM_CLOSE,
END_RESPONSE_TABLE;

//****************************************************************

TicketsWindow::TicketsWindow(Registry *reg)
	: WXFrameWindow(reg, 0, ResString((TModule *)0, IDS_TICKETW),
		new TListWindow(0, 1, 0, 0, 0, 0), TRUE)
{
	lw = dynamic_cast<TListWindow *>(ClientWnd);
	lw->Attr.Style |= LVS_REPORT;

	ZeroMemory(&TrayData, sizeof(TrayData));
	TrayData.cbSize = sizeof(TrayData);
}

TicketsWindow::~TicketsWindow()
{
	if (TrayData.uFlags)
		Shell_NotifyIcon(NIM_DELETE, &TrayData);

}

char far*
TicketsWindow::GetClassName()
{
	return TICKET_WINDOW_CLASS;
}

bool
TicketsWindow::Create()
{
	SetIcon(GetApplication(), IDI_TICKETS);
	SetIconSm(GetApplication(), IDI_TICKETS);
	AssignMenu(IDM_TICKETS);
	int s = SetupAttr();
	Attr.AccelTable = IDM_TICKETS;

	if ((!SetupMgr.NormalWindow && !SetupMgr.SysTray) ||
	    (SetupMgr.NormalWindow))
		Attr.Style |= WS_VISIBLE;
	else {
		Attr.Style &= ~WS_VISIBLE;
		s = SW_HIDE;
		}
	if (IsFlagSet(wfMainWindow))
		GetApplication()->SetCmdShow(s);

	bool res = WXFrameWindow::Create();

	if (res && !IsFlagSet(wfMainWindow)) {
		if ((s == SW_HIDE) && (TrayData.uFlags == 0))
			s = SW_RESTORE;
		ShowWindow(s);
		}

	return res;
}

void
TicketsWindow::SetupWindow()
{
	char time_str[64];
	time_t t;

	WXFrameWindow::SetupWindow();

	if (SetupMgr.SysTray) {
		TrayData.hWnd = *this;
		TrayData.uFlags = NIF_ICON|NIF_TIP|NIF_MESSAGE;
		TrayData.uCallbackMessage =
			::RegisterWindowMessage(WM_REG_TRAYNOTE);
		TrayData.hIcon = GetApplication()->
			LoadIcon(MAKEINTRESOURCE(IDI_TICKETS));
		strlcpy(TrayData.szTip, "Kerberos Ticket Manager",
			sizeof(TrayData.szTip));
		if (!Shell_NotifyIcon(NIM_ADD, &TrayData)) {
			TrayData.uFlags = 0;
			ShowWindow(SW_RESTORE);
			if (IsFlagSet(wfMainWindow))
				GetApplication()->SetCmdShow(SW_RESTORE);
			}
		}

	// Don't try to give localtime an invalid time, it might
	// return NULL then.... Tanks to Nick Lloyd <lloyd@irg.usask.ca>
	// who helped me pinpoint this problem.
	t = time_r(NULL);
	struct tm *tm = localtime(&t);
	strftime(time_str, sizeof(time_str), "%c", tm);
	int time_width = lw->GetStringWidth(time_str) + 30;

	lw->InsertColumn(0, TListWindColumn("Principal",
		lw->GetStringWidth("krbtgt.STACKEN.KTH.SE@STACKEN.KTH.SE") + 30));
	lw->InsertColumn(1, TListWindColumn("Start time", time_width));
	lw->InsertColumn(2, TListWindColumn("End time", time_width));
	lw->InsertColumn(3, TListWindColumn("Kvno", 
		lw->GetStringWidth("KvnoXXXXXXXX") + 15));
	lw->InsertColumn(4, TListWindColumn("User",
		lw->GetStringWidth("XXXXXXX.XXXXXXXX") + 15));

	UpdateCache();
}

void
TicketsWindow::Destroy(int retVal)
{
	TWindow::Destroy(retVal);
}


LRESULT
TicketsWindow::EvUpdateCache(WPARAM, LPARAM)
{
	UpdateCache();
	return 0;
}


#pragma argsused
LRESULT
TicketsWindow::TrayNotify(WPARAM wParam, LPARAM lParam)
{
	if (lParam == WM_RBUTTONUP) {
		POINT pnt;
		if (!::GetCursorPos(&pnt)) {
			pnt.x = GetSystemMetrics(SM_CXSCREEN);
			pnt.y = GetSystemMetrics(SM_CYSCREEN);
			}
		TPopupMenu(TMenu(*GetApplication(), IDM_TRAYPOPUP)).
			TrackPopupMenu(TPM_RIGHTALIGN|TPM_RIGHTBUTTON,
				pnt.x, pnt.y, 0, *this);
		}
	else if (lParam == WM_LBUTTONUP)
		::PostMessage(*this, WM_COMMAND, CM_SHOWTICKETS, 0);
	return 0;
}

#pragma argsused
void
TicketsWindow::EvActivate(uint active, bool minimized, HWND hWndOther)
{
	UpdateCache();
}

void
TicketsWindow::EvClose()
{
	if (TrayData.uFlags) {
		ShowWindow(SW_HIDE);
		return;
		}
	WXFrameWindow::EvClose();
}

void
TicketsWindow::CmExit()
{
	WXFrameWindow::EvClose();
}

void
TicketsWindow::CmLogin()
{
	KClientKey k;

	KClient kc(this, &SetupData, 0, 0, 0, 0);
	kc.SetDefaultPrincipal("");
	kc.Login(&k);
}

void
TicketsWindow::CmLogin5()
{
	k5c->Login(this, 0);
}

void
TicketsWindow::CmRefresh()
{
	UpdateCache();
}

void
TicketsWindow::UpdateCache()
{
	bool TicketsFound = false;

	lw->DeleteAllItems();

	strlcpy(NewTitle, ResString(this, IDS_NOTICKETCUE), sizeof(NewTitle));

	UpdateCache4(TicketsFound);
	UpdateCache5(TicketsFound);

	SetCaption(NewTitle);
}

void
TicketsWindow::UpdateCache4(bool & TicketsFound)
{
	register int num = 0;
	time_t end;
	CREDENTIALS c;
	struct tm  *tm;
	char Buffer[512];

	if (tf_init(NULL, R_TKT_FIL) ||
	    tf_get_pname(Buffer))
		return;

	while (tf_get_cred(&c) == KSUCCESS) {
		if (num == 0) {
			if (TicketsFound) {
				TListWindItem item("------------");
				item.SetIndex(lw->GetItemCount());
				lw->InsertItem(item);
				}
			strlcpy(prealm, c.realm, sizeof(prealm));
			strlcpy(pname, c.pname, sizeof(pname));
			strlcpy(pinst, c.pinst, sizeof(pinst));
			strlcpy(NewTitle, 
				krb_unparse_name_long(pname, pinst, prealm),
				sizeof(NewTitle));
			}
		TicketsFound = true;
		end = krb_life_to_time(c.issue_date, c.lifetime);

		TListWindItem item(krb_unparse_name_long(c.service,
			c.instance, c.realm));

		item.SetIndex(lw->GetItemCount());
		lw->InsertItem(item);

		if ((tm = localtime((time_t*)&c.issue_date)) == NULL)
			strlcpy(Buffer, "?????????", sizeof(Buffer));
		else
			strftime(Buffer, sizeof(Buffer), "%c", tm);
		item.SetText(Buffer);
		item.SetSubItem(1);
		lw->SetItem(item);

		if ((tm = localtime(&end)) == NULL)
			strlcpy(Buffer, "?????????", sizeof(Buffer));
		else
			strftime(Buffer, sizeof(Buffer), "%c", tm);

		item.SetSubItem(2);
		lw->SetItem(item);

		item.SetText(itoa(c.kvno, Buffer, 10), 3);
		item.SetSubItem(3);		
		lw->SetItem(item);

		strlcpy(Buffer, c.pname, sizeof(Buffer));
		if (c.pinst[0]) {
			strlcat(Buffer, ".", sizeof(Buffer));
			strlcat(Buffer, c.pinst, sizeof(Buffer));
			}

		item.SetText(Buffer);
		item.SetSubItem(4);
		lw->SetItem(item);

		num++;
		}

	if (krb_get_config_bool("nat_in_use")) {
		NAT_REALM_ID nri;

		TListWindItem item("------------");
		item.SetIndex(lw->GetItemCount());
		lw->InsertItem(item);
		item.SetSubItem(1);
		lw->SetItem(item);
		item.SetSubItem(2);
		lw->SetItem(item);
		item.SetSubItem(4);
		lw->SetItem(item);

		item.SetText("--");
		item.SetSubItem(3);
		lw->SetItem(item);

		item.SetText(Buffer);
		item.SetSubItem(0);

		for (num = 0;
		     tf_get_nat_struct(num, &nri, sizeof(nri)) == KSUCCESS;
		     num++)  {

			strlcpy(Buffer,	inet_ntoa(nri.ip_addr),
				sizeof(Buffer));
			strlcat(Buffer, " for ", sizeof(Buffer));
			strlcat(Buffer, nri.realm, sizeof(Buffer));
			item.SetIndex(lw->GetItemCount());
			lw->InsertItem(item);
			}
		}
	

	return;
}

void
TicketsWindow::UpdateCache5(bool & TicketsFound)
{
	krb5_principal principal;
	krb5_cc_cursor cursor;
	krb5_creds creds;
	bool first = true;

	
	if (!(bool)*k5c)
		return;

//	krb5_set_default_realm(context, "");

	if (!k5c->CCGetPrincipal(&principal))
		return;

	if (!k5c->CCStartSeqGet(&cursor))
		return;

	while (k5c->CCNextCred(&creds, &cursor)) {
		char *str;
		Ticket t;
		size_t len;
		char buf[128], buf2[32];
		int ret;

		if (!k5c->UnparseName(creds.server, &str))
			continue;

		if (first) {
			char *s;
			if (TicketsFound) {
				TListWindItem item("------------");
				item.SetIndex(lw->GetItemCount());
				lw->InsertItem(item);
				}
			if (k5c->UnparseName(principal, &s)) {
				strlcpy(NewTitle, s, sizeof(NewTitle));
				free(s);
				}
			}

		first = false;
		TicketsFound = true;

		TListWindItem item(str);
		item.SetIndex(lw->GetItemCount());
		lw->InsertItem(item);
		free (str);

		item.SetText(printable_time(
			creds.times.starttime ? creds.times.starttime
				: creds.times.authtime));
		item.SetSubItem(1);
		lw->SetItem(item);

		item.SetText(printable_time(creds.times.endtime));
		item.SetSubItem(2);
		lw->SetItem(item);

		decode_Ticket((const unsigned char *)creds.ticket.data,
			creds.ticket.length, &t, &len);
		ret = krb5_enctype_to_string(*k5c,
			(krb5_enctype)t.enc_part.etype, &str);
		if (ret == 0) {
			strlcpy(buf, str, sizeof(buf));
			free(str);
			}
		else {
			snprintf(buf, sizeof(buf), "unknown(%d)",
				t.enc_part.etype);
			}
		if (t.enc_part.kvno) {
			snprintf(buf2, sizeof(buf2), "/%d",
				*t.enc_part.kvno);
			strlcat(buf, buf2, sizeof(buf));
			}
		if (creds.session.keytype != t.enc_part.etype) {
			ret = krb5_keytype_to_string(*k5c,
				(krb5_keytype)creds.session.keytype, &str);
			if (ret == KRB5_PROG_KEYTYPE_NOSUPP)
				ret = krb5_enctype_to_string(*k5c,
					(krb5_enctype)creds.session.keytype,
					&str);
			if (ret)
				strlcat(buf, "/???", sizeof(buf));
			else {
				strlcat(buf, "/", sizeof(buf));
				strlcat(buf, str, sizeof(buf));
				free(str);
				}
			}
		free_Ticket(&t);

		item.SetText(buf);
		item.SetSubItem(3);
		lw->SetItem(item);

		if (k5c->UnparseNameUserInst(creds.client, &str)) {
			item.SetText(str);
			item.SetSubItem(4);
			lw->SetItem(item);
			free (str);
			}

		k5c->FreeCredsContents(&creds);
		}
	k5c->CCEndSeqGet(&cursor);
	k5c->FreePrincipal(principal);
	return;
}

void
TicketsWindow::CmClear()
{
	if (MessageBox(ResString(this, IDS_RCLEAR), header,
		MB_YESNO|MB_ICONQUESTION) != IDYES)
		return;			

	dest_tkt();

	k5c->CCDestroy();

	UpdateCache();
}

void
TicketsWindow::CmTest()
{
	krb5_creds this_cred, *cred;
	char princ_str[128];

	cred = 0;
	memset(&this_cred, 0, sizeof(this_cred));
	princ_str[0] = 0;
	if (TInputDialog(this, "TEST", "Kerberos 5 principal",
		princ_str, sizeof(princ_str)).Execute() != IDOK)
		return;

	if (!k5c->ParseName(princ_str, &this_cred.server)) {
		MessageBox(k5c->GetErrorText(), "ParseName returns",
			MB_ICONEXCLAMATION|MB_OK);
		return;
		}

	if (!k5c->MakeRequestLogin(this, this_cred.server)) {
		MessageBox(k5c->GetErrorText(), "MakeRequestLogin",
			MB_ICONEXCLAMATION|MB_OK);
		k5c->FreeCredsContents(&this_cred);
		return;
		}

	if (!k5c->CCGetPrincipal(&this_cred.client)) {
		MessageBox(k5c->GetErrorText(), "CCGetPrincipal",
			MB_ICONEXCLAMATION|MB_OK);
		k5c->FreeCredsContents(&this_cred);
		return;
		}

	this_cred.times.endtime = 0;
	if (k5c->GetAuthContext()->keytype)
		this_cred.session.keytype = k5c->GetAuthContext()->keytype;

	if (!k5c->GetCredentials(&this_cred, &cred)) {
		MessageBox(k5c->GetErrorText(), "GetCredentials",
			MB_ICONEXCLAMATION|MB_OK);
		k5c->FreeCredsContents(&this_cred);
		return;
		}
	krb5_free_creds(*k5c, cred);
	k5c->FreeCredsContents(&this_cred);
}

void
TicketsWindow::CmCPW4()
{
	ChangePasswordK4(this).Execute();
}

void
TicketsWindow::CmCPW5()
{
	ChangePasswordK5(this).Execute();
}

//****************************************************************

ChangePassword::ChangePassword(TWindow *parent, int captionId)
	: TDialog(parent, IDD_KRBCPW)
{
	new TEdit(this, IDC_CPWUSER, CPW_MAXSTR);
	new TEdit(this, IDC_CPWINST, INST_SZ);
	new TEdit(this, IDC_CPWREALM, REALM_SZ);
	new TEdit(this, IDC_CPWOLD, CPW_MAXSTR);
	new TEdit(this, IDC_CPWNEW, CPW_MAXSTR);
	new TEdit(this, IDC_CPWVER, CPW_MAXSTR);

	SetTransferBuffer(&trfbuf);
	ZeroMemory(&trfbuf, sizeof(trfbuf));
	OldCur = 0;
	CaptionId = captionId;
}

ChangePassword::~ChangePassword()
{
	ZeroMemory(&trfbuf, sizeof(trfbuf));
}

void
ChangePassword::SetupWindow()
{
	TDialog::SetupWindow();
	if (CaptionId != 0)
		SetCaption(ResString(this, CaptionId));
}

void
ChangePassword::CloseWindow(int code)
{
	const char *expl;

	if (code != IDOK) {
		TDialog::CloseWindow(code);
		return;
		}

	TransferData(tdGetData);

	OldCur = ::SetCursor(::LoadCursor(0, IDC_WAIT));

	expl = CheckPassword(trfbuf.pnew);
	if (expl) {
		::SetCursor(OldCur);
		OldCur = 0;
		if (MessageBox(expl, errorHeader,
			MB_OKCANCEL|MB_ICONEXCLAMATION) == IDCANCEL)
			TDialog::CloseWindow(IDCANCEL);
		return;
		}

	if (strcmp(trfbuf.pnew, trfbuf.ver) != 0) {
		ErrorS(IDS_PSWNEQ);
		return;
		}

	expl = DoPSWChange();

	if (OldCur) {
		::SetCursor(OldCur);
		OldCur = 0;
		}

	if (expl) {
		MessageBox(expl, header, MB_OK);
		free((void *)expl);
		}
}

void
ChangePassword::ErrorS(int ResId, const char *kerr, const char *note)
{
	char buf[512];

	ZeroMemory(&trfbuf, sizeof(trfbuf));

	if (OldCur) {
		::SetCursor(OldCur);
		OldCur = 0;
		}

	strlcpy(buf, ResString(this, ResId), sizeof(buf));
	if (kerr) {
		strlcat(buf, "\n", sizeof(buf));
		strlcat(buf, kerr, sizeof(buf));
		}

	if (note) {
		if (strncmp(KADM_ULOSE, note, strlen(KADM_ULOSE)) == 0) {
			strlcat(buf, ResString(this, IDS_PSWULOSE),
				sizeof(buf));
			}
		else {
			strlcat(buf, "\n", sizeof(buf));
			strlcat(buf, note, sizeof(buf));
			}
		}

	if (MessageBox(buf, errorHeader,
		MB_OKCANCEL|MB_ICONEXCLAMATION) == IDCANCEL)
		TDialog::CloseWindow(IDCANCEL);
}

//****************************************************************

ChangePasswordK4::ChangePasswordK4(TWindow *parent)
	: ChangePassword(parent, CM_KRBCPW4)
{
	snprintf(SpecTktFile, sizeof(SpecTktFile), "krb_cpw_%08x",
		GetApplication()->GetInstance());

	SavedTicketFile = wkrb_new_tktfile(SpecTktFile);
	if (tf_get_pinst(trfbuf.inst) != KSUCCESS)
		trfbuf.inst[0] = 0;
	krb_get_lrealm(trfbuf.realm, 1);
}

ChangePasswordK4::~ChangePasswordK4()
{
	if (SavedTicketFile)
		wkrb_restore_tktfile(SavedTicketFile);
}

const char *
ChangePasswordK4::CheckPassword(char *pword)
{
	int ret = kadm_check_pw(pword);

	if (!ret)
		return NULL;
	
	return wkadm_get_error(ret);
}

char *
ChangePasswordK4::DoPSWChange()
{
	char buf[1024];
	int status;
	des_cblock newkey;
	char *pw_msg;

	if (krb_get_config_bool("nat_in_use"))
		status = krb_get_pw_in_tkt(trfbuf.user, trfbuf.inst,
			trfbuf.realm, "krbtgt", trfbuf.realm, 1, trfbuf.old);
	else
		status = KSUCCESS;

	if (status == KSUCCESS)
		status = krb_get_pw_in_tkt(trfbuf.user, trfbuf.inst,
			trfbuf.realm, PWSERV_NAME, KADM_SINST, 1, trfbuf.old);

	if (status == INTK_BADPW) {
		Error(IDS_BADPSW);
		return NULL;
		}

	if (status != KSUCCESS) {
		Error(IDS_UNKKERR, status + IDS_PREF_KERR);
		return NULL;
		}

	status = kadm_init_link (PWSERV_NAME, KRB_MASTER, trfbuf.realm);
	if (status != KADM_SUCCESS) {
		Error(IDS_WINIT);
		return NULL;
		}

	des_string_to_key(trfbuf.pnew, &newkey);
	status = kadm_change_pw_plain((unsigned char*)&newkey,
		trfbuf.pnew, &pw_msg);
	ZeroMemory(newkey, sizeof(newkey));

	if (status != KADM_SUCCESS) {
		Error(IDS_UNKKERR, status, pw_msg);
		return NULL;
		}

	strlcpy(buf, ResString(this, IDS_PSWCOK), sizeof(buf));
	strlcat(buf, trfbuf.user, sizeof(buf));
	if (trfbuf.inst[0]) {
		strlcat(buf, ".", sizeof(buf));
		strlcat(buf, trfbuf.inst, sizeof(buf));
		}
	strlcat(buf, "@", sizeof(buf));
	strlcat(buf, trfbuf.realm, sizeof(buf));

	TDialog::CloseWindow(IDOK);
	ZeroMemory(&trfbuf, sizeof(trfbuf));
	return strdup(buf);
}

void
ChangePasswordK4::Error(int ResId, int kerr, const char *note)
{
	dest_tkt();
	ErrorS(ResId, wkadm_get_error(kerr), note);
}

//****************************************************************

ChangePasswordK5::ChangePasswordK5(TWindow *parent)
	: ChangePassword(parent, CM_KRBCPW5)
{
	krb5_principal princ;

	k5c = new Kerberos5Context(&SetupData);

	if (k5c->CCGetPrincipal(&princ)) {
		char *cp, *c;
		if (k5c->UnparseNameUserInst(princ, &cp)) {
			if ((c = strchr(cp, '/')) != NULL) {
				*c++ = 0;
				strlcpy(trfbuf.inst, c, sizeof(trfbuf.inst));
				}
			strlcpy(trfbuf.user, cp, sizeof(trfbuf.user));
			free(cp);
			}

		strlcpy(trfbuf.realm, *krb5_princ_realm(*k5c, princ),
			sizeof(trfbuf.realm));
		}
	else {
		DWORD un = sizeof(trfbuf.user)-1;

		GetUserName(trfbuf.user, &un);
		krb5_realm *realms = krb5_config_get_strings(*k5c, NULL,
			"libdefaults", "default_realm", NULL);
		if (*realms)
			strlcpy(trfbuf.realm, *realms, sizeof(trfbuf.realm));
		if (realms)
			krb5_free_host_realm(*k5c, realms);
		}
}

ChangePasswordK5::~ChangePasswordK5()
{
	delete k5c;
}

#pragma argsused
const char *
ChangePasswordK5::CheckPassword(char *pword)
{
	return NULL;
}

char *
ChangePasswordK5::DoPSWChange()
{
	char buf[1024];
	krb5_principal principal;
	krb5_error_code ret;
	krb5_creds cred;
	krb5_get_init_creds_opt opt;
	krb5_data result_code_string, result_string;
	int result_code;
	register int i;
	char *cp;

	k5c->OpenLog("Password change");

	krb5_get_init_creds_opt_init (&opt);
	krb5_get_init_creds_opt_set_tkt_life (&opt, 300);
	krb5_get_init_creds_opt_set_forwardable (&opt, FALSE);
	krb5_get_init_creds_opt_set_proxiable (&opt, FALSE);

	if (!k5c->Conv4to5Principal(KCPBuf(trfbuf.user, trfbuf.inst,
					  trfbuf.realm, 0),
					  &principal)) {
		Error(IDS_UNKKERR, k5c->ret);
		return NULL;
		}

	ret = krb5_get_init_creds_password(*k5c,
					&cred,
					principal,
					trfbuf.old,
					krb5_prompter_posix,
					NULL,
					0,
					"kadmin/changepw",
					&opt);
	switch (ret) {
		case 0:
			break;
		case KRB5KRB_AP_ERR_BAD_INTEGRITY :
		case KRB5KRB_AP_ERR_MODIFIED :
			Error(IDS_BADPSW, ret);
			return NULL;
		default:
			Error(IDS_UNKKERR, ret,
			      "krb5_get_init_creds_password");
			return NULL;
		}

	krb5_data_zero(&result_code_string);
	krb5_data_zero(&result_string);

	ret = krb5_change_password(*k5c, &cred, trfbuf.pnew,
				   &result_code,
				   &result_code_string,
				   &result_string);
	if (ret) {
		Error(IDS_UNKKERR, ret, "krb5_change_password");
		krb5_free_creds_contents (*k5c, &cred);
		return NULL;
		}

	strlcpy(buf, ResString(this, IDS_K5REPLY), sizeof(buf));
	i = (int)result_string.length;
	if (i > (sizeof(buf) - strlen(buf) - 1))
		i = (sizeof(buf) - strlen(buf) - 1);
	
	cp = &buf[strlen(buf)];
	CopyMemory(cp, result_string.data, i);
	cp[i] = 0;

	krb5_data_free (&result_code_string);
	krb5_data_free (&result_string);
    
	krb5_free_creds_contents (*k5c, &cred);

	TDialog::CloseWindow(IDOK);
	return strdup(buf);
}

void
ChangePasswordK5::Error(int ResId, int kerr, const char *note)
{
	ErrorS(ResId, wkadm_get_error(kerr), note);
}

