/*
 * Copyright (c) 2003 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden). 
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 */

#include "tsasl_locl.h"
#include <gssapi.h>

RCSID("$Id$");

struct tsasl_gss {
    int tg_flags;
#define TM_F_INT	1
#define TM_F_CONF	2
    int tg_state;
    gss_ctx_id_t tg_hdl;
    gss_name_t tg_peer_name;
    int tg_my_max_size;
    int tg_peer_max_size;
    int tg_neg_sec_bits;
    
    struct tsasl_buffer tg_in;
    unsigned char tg_hdr[4];
    size_t tg_pkt_off;
};


static int tsasl_gss_max_secbits = 0x1 + 0x2 + 0x4;

#if 1
#define	mech_oid	GSS_KRB5_MECHANISM
#else
#define	mech_oid	GSS_SPNEGO_MECHANISM
#endif

static void
set_gss_errorstring(struct tsasl_peer *peer, 
		    OM_uint32 min_stat,
		    const char *fmt, ...)
{
    OM_uint32 new_stat;
    OM_uint32 msg_ctx = 0;
    gss_buffer_desc status_string;
    OM_uint32 ret;
    char *msg1, *msg2;
    va_list ap;

    va_start(ap, fmt);
    vasprintf(&msg1, fmt, ap);
    va_end(ap);

    do {
	ret = gss_display_status (&new_stat,
				  min_stat,
				  GSS_C_MECH_CODE,
				  GSS_C_NO_OID,
				  &msg_ctx,
				  &status_string);
	asprintf (&msg2, "%s %s\n", msg1, (char *)status_string.value);
	gss_release_buffer (&new_stat, &status_string);
	free(msg1);
	msg1 = msg2;
    } while (!GSS_ERROR(ret) && msg_ctx != 0);

    _tsasl_set_errorstring(peer,"%s", msg1);
}

static int
parse_sec_bits(struct tsasl_peer *peer, 
	       unsigned char *buf, int sz, int *bits)
{
    struct tsasl_gss *tg;
	
    tg = peer->tp_mechprivate;
	
    if (sz < 4) {
	_tsasl_set_errorstring(peer, "gss: security bits too short");
	return TSASL_ERROR;
    }
    *bits = buf[0] & tsasl_gss_max_secbits;
    tg->tg_peer_max_size = (buf[1] << 16) + (buf[2] << 8) + buf[3];
    return 0;
}

static int
neg_best_sec_bits(struct tsasl_peer *peer, int bits)
{
    struct tsasl_gss *tg;
    int error;

    tg = peer->tp_mechprivate;

    error = 0;

    if (bits & 4) {
	bits = 4;
	peer->tp_ctxflags |= TP_CTX_FLAGS_CONF | TP_CTX_FLAGS_INT;
    } else if (peer->tp_flags & TSASL_FLAGS_CONFIDENTIALITY) {
	_tsasl_set_errorstring(peer, "gss: req conf, not supported by peer");
	error = TSASL_ERROR;
    } else if (bits & 2) {
	bits = 2;
	peer->tp_ctxflags |= TP_CTX_FLAGS_INT;
    } else if (peer->tp_flags & TSASL_FLAGS_INTEGRITY) {
	_tsasl_set_errorstring(peer, "gss: req int, not supported by peer");
	error = TSASL_ERROR;
    } else if (bits & 1) {
	bits = 1;
    } else {
	_tsasl_set_errorstring(peer, "gss: no sec bit set");
	error = TSASL_ERROR;
    }
    if (error == 0)
	tg->tg_neg_sec_bits = bits;
    return error;
}

static int
unparse_bit_msg(struct tsasl_peer *peer, gss_buffer_t input_token,
		gss_buffer_t output_token, int *bits)
{
    OM_uint32 maj_stat, min_stat;
    struct tsasl_gss *tg;
    gss_qop_t qop_state;
    int conf_state;
    int error;
		
    tg = peer->tp_mechprivate;

    maj_stat = gss_unwrap(&min_stat,
			  tg->tg_hdl,
			  input_token,
			  output_token,
			  &conf_state,
			  &qop_state);
    if (GSS_ERROR(maj_stat)) {
	set_gss_errorstring(peer, min_stat,
			    "gss: sec bit exchange with conf set");
	output_token->length = 0;
	return TSASL_ERROR;
    }
	
    if (conf_state != 0) {
	_tsasl_set_errorstring(peer, "gss: sec bit exchange "
			      "with conf set");
	if (output_token->length)
	    maj_stat = gss_release_buffer(&min_stat,
					  output_token);
	return TSASL_ERROR;
    }

    error = parse_sec_bits(peer,
			   output_token->value,
			   output_token->length,
			   bits);
    if (error && output_token->length)
	maj_stat = gss_release_buffer(&min_stat, output_token);
    return error;
}


static int
gss_auth_initiator(struct tsasl_peer *peer, struct tsasl_gss *tg,
		   int flags,
		   struct tsasl_buffer *in,
		   struct tsasl_buffer *out)
{
    OM_uint32 maj_stat, min_stat;
    gss_buffer_desc input_token, output_token;
    int error;

    input_token.length = in->tb_size;
    input_token.value = in->tb_data;

    output_token.length = 0;

    switch(tg->tg_state) {
    case 0:
    case 1:
	maj_stat = 
	    gss_init_sec_context(&min_stat,
				 GSS_C_NO_CREDENTIAL,
				 &tg->tg_hdl,
				 tg->tg_peer_name,
				 mech_oid,
				 GSS_C_MUTUAL_FLAG | 
				 GSS_C_SEQUENCE_FLAG,
				 0,
				 GSS_C_NO_CHANNEL_BINDINGS,
				 &input_token,
				 NULL,
				 &output_token,
				 NULL,
				 NULL);

	if (GSS_ERROR(maj_stat)) {
	    set_gss_errorstring(peer, min_stat,	"gss: init context error");
	    error = TSASL_ERROR;
	} else if (maj_stat & GSS_S_CONTINUE_NEEDED) {
	    error = TSASL_CONTINUE;
	} else {
	    error = TSASL_CONTINUE;
	    tg->tg_state = 2;
	}

	break;
    case 2: {
	char *authzid;
	unsigned char *buf;
	int authzidsz;
	int bits;
	int conf_state;
		
	error = unparse_bit_msg(peer, 
				&input_token,
				&output_token,
				&bits);
	if (error == 0) {
	    if (output_token.length != 4) {
		_tsasl_set_errorstring(peer, "gss: "
				       "server sec token too large");
		error = TSASL_ERROR;
	    }
	    gss_release_buffer(&min_stat, &output_token);
	}
	if (error)
	    break;

	error = neg_best_sec_bits(peer, bits);
	if (error)
	    break;

	error = tsasl_authorization_id_get(peer, &authzid);
	if (error == 0) {
	    authzidsz = strlen(authzid);
	} else if (error == ENOENT) {
	    authzidsz = 0;
	    authzid = NULL;
	} else {
	    _tsasl_set_errorstring(peer, "gss: can't get authzid");
	    error = TSASL_ERROR;
	    break;
	}

	if ((buf = malloc(authzidsz + 4)) == NULL) {
	    _tsasl_set_errorstring(peer, "gss: out of memory");
	    error = TSASL_ERROR;
	    if (authzid)
		tsasl_authorization_id_free(authzid);
	    break;
	}
		
	buf[0] = tg->tg_neg_sec_bits;
	buf[1] = (tg->tg_my_max_size >> 16) & 0xff;
	buf[2] = (tg->tg_my_max_size >> 8) & 0xff;
	buf[3] = (tg->tg_my_max_size) & 0xff;
	memcpy(buf + 4, authzid, authzidsz);
	if (authzid)
	    tsasl_authorization_id_free(authzid);
			
	input_token.length = authzidsz + 4;
	input_token.value = buf;

	maj_stat = gss_wrap(&min_stat, tg->tg_hdl,
			    0, /* conf_req_flag */
			    GSS_C_QOP_DEFAULT,
			    &input_token,
			    &conf_state,
			    &output_token);
	if (GSS_ERROR(maj_stat)) {
	    set_gss_errorstring(peer, min_stat,
				"gss: wrap sec bit error");
	    error = TSASL_ERROR;
	}
			
	peer->tp_ctxflags |= TP_CTX_FLAGS_AUTHDONE;
	tg->tg_state = 3;
	error = TSASL_DONE;
	break;
    }
    case 3:
    default:
	abort();
    }

    if (output_token.length && 
	(error == TSASL_DONE || error == TSASL_CONTINUE))
    {
	int error2;
	error2 = _tsasl_buffer_alloc(out, output_token.length);
	if (error2) {
	    gss_release_buffer(&min_stat, &output_token);
	    return error2;
	}
	memcpy(out->tb_data, output_token.value, out->tb_size);
    } else
	_tsasl_buffer_zero(out);

    if (output_token.length)
	gss_release_buffer(&min_stat, &output_token);

    return error;
}

static int
gss_auth_acceptor(struct tsasl_peer *peer, struct tsasl_gss *tg,
		  int flags,
		  struct tsasl_buffer *in,
		  struct tsasl_buffer *out)
{
    OM_uint32 maj_stat, min_stat;
    gss_buffer_desc input_token, output_token;
    int error;

    input_token.length = in->tb_size;
    input_token.value = in->tb_data;

    output_token.length = 0;

    switch(tg->tg_state) {
    case 0:
	if (input_token.length == 0) {
	    error = TSASL_CONTINUE;
	    tg->tg_state = 1;
	    break;
	}
    case 1:
	maj_stat = gss_accept_sec_context (&min_stat,
					   &tg->tg_hdl,
					   GSS_C_NO_CREDENTIAL,
					   &input_token,
					   GSS_C_NO_CHANNEL_BINDINGS,
					   &tg->tg_peer_name,
					   NULL,
					   &output_token,
					   NULL,
					   NULL,
					   NULL);

		
	if (GSS_ERROR(maj_stat)) {
	    set_gss_errorstring(peer, min_stat, "gss: accept_sec_context");
	    error = TSASL_ERROR;
	} else if (maj_stat & GSS_S_CONTINUE_NEEDED) {
	    error = TSASL_CONTINUE;
	} else {
	    error = TSASL_CONTINUE;
	    tg->tg_state = 2;
	}
		
	/*
	 * if there was an output token, send it
	 * if we still are in state 1, continue to be so
	 * else just fall though to state 2.
	 */
	if (output_token.length) {
	    break;
	} else if (tg->tg_state == 1)
	    break;
	/* FALLTHROUGH */
    case 2: {
	char buf[4];
	int conf_state;
	/* send gss_wraped security flags */
		
	if (input_token.length != 0) {
	    _tsasl_set_errorstring(peer, "gss: "
				   "client didn't sent empty token");
	    error = TSASL_ERROR;
	    break;
	}

	if (peer->tp_flags & TSASL_FLAGS_CONFIDENTIALITY)
	    buf[0] = 4;
	else if (peer->tp_flags & TSASL_FLAGS_INTEGRITY)
	    buf[0] = 2 + 4;
	else
	    buf[0] = 1 + 2 + 4;
	buf[1] = (tg->tg_my_max_size >> 16) & 0xff;
	buf[2] = (tg->tg_my_max_size >> 8) & 0xff;
	buf[3] = (tg->tg_my_max_size) & 0xff;

	input_token.value = buf;
	input_token.length = 4;
	output_token.length = 0;

	maj_stat = gss_wrap(&min_stat, 
			    tg->tg_hdl,
			    0, /* conf_req_flag */
			    GSS_C_QOP_DEFAULT,
			    &input_token,
			    &conf_state,
			    &output_token);
	if (GSS_ERROR(maj_stat)) {
	    set_gss_errorstring(peer, min_stat, "gss: wrap init sec bits");
	    error = TSASL_ERROR;
	} else {
	    tg->tg_state = 3;
	    error = TSASL_CONTINUE;
	}

	break;
    }
    case 3: {
	int bits;

	error = unparse_bit_msg(peer, 
				&input_token,
				&output_token,
				&bits);
	if (error)
	    break;

	if (output_token.length > tg->tg_my_max_size) {
	    _tsasl_set_errorstring(peer, "authzid too large");
	    error = TSASL_ERROR;
	    break;
	}

	/* verify bits */
	if ((bits & 4) == 0 &&
	    (peer->tp_flags & TSASL_FLAGS_CONFIDENTIALITY)) {
	    _tsasl_set_errorstring(peer, "gss:req conf, "
				   "not supported by peer");
	    error = TSASL_ERROR;
	} else if (bits & 4) {
	    ;
	} else if ((bits & 2) == 0 &&
		   (peer->tp_flags & TSASL_FLAGS_INTEGRITY)) {
	    _tsasl_set_errorstring(peer, "gss: req int, "
				   "not supported by peer");
	    error = TSASL_ERROR;
	} else if (bits & 2) {
	    ;
	}
	if (error)
	    break;

	if (bits == 4) {
	    peer->tp_ctxflags |= TP_CTX_FLAGS_CONF | TP_CTX_FLAGS_INT;
	} else if (bits == 2) {
	    peer->tp_ctxflags |= TP_CTX_FLAGS_INT;
	} else if (bits == 1) {
	    ;
	} else if (bits) { 
	    _tsasl_set_errorstring(peer, "gss: too many sec bit set");
	    error = TSASL_ERROR;
	    break;
	} else {
	    _tsasl_set_errorstring(peer, "gss: no sec bit set");
	    error = TSASL_ERROR;
	    break;			
	}
		
	tg->tg_neg_sec_bits = bits;

	if (output_token.length > 4) {
	    char *buf;
	    int len;

	    len = output_token.length - 4 + 1;

	    if ((buf = malloc(len)) == NULL) {
		_tsasl_set_errorstring(peer, "gss: out of mem");
		error = TSASL_ERROR;
		break;
	    }

	    memcpy(buf, (char *)output_token.value + 4, len);
	    buf[len - 1] = '\0';

	    error = tsasl_authorization_id_set(peer, buf);
	    free(buf);

	    if (error)
		_tsasl_set_errorstring(peer, "gss: failed setting authzid");
	}

	if (error == 0) {
	    char *exported_name, *name;
	    int len;

	    error = TSASL_DONE;
	    tg->tg_state = 4;
	    peer->tp_ctxflags |= TP_CTX_FLAGS_AUTHDONE;

	    if (output_token.length) 
		maj_stat = gss_release_buffer(&min_stat, &output_token);
	    output_token.length = 0;

	    maj_stat = gss_export_name(&min_stat, 
				       tg->tg_peer_name,
				       &output_token);
	    if (GSS_ERROR(maj_stat)) {
		set_gss_errorstring(peer, min_stat, "gss: export name");
		break;
	    }

	    len = base64_encode(output_token.value, 
				output_token.length,
				&exported_name);
	    maj_stat = gss_release_buffer(&min_stat,&output_token);

	    if (len <= 0) {
		free(exported_name);
		error = TSASL_ERROR;
		break;
	    }

	    asprintf(&name, "GSSAPI:%s", exported_name);
	    free(exported_name);
	    tsasl_authentication_id_set(peer, name);
	    free(name);
	}

	break;
    }
    case 4:
    default:
	abort();
    }

    if (output_token.length && 
	(error == TSASL_DONE || error == TSASL_CONTINUE))
    {
	int error2;

	error2 = _tsasl_buffer_alloc(out, output_token.length);
	if (error2) {
	    maj_stat = gss_release_buffer(&min_stat, &output_token);
	    return error2;
	}
	memcpy(out->tb_data, output_token.value, out->tb_size);
	maj_stat = gss_release_buffer(&min_stat, &output_token);
    } else if (output_token.length) 
	maj_stat = gss_release_buffer(&min_stat, &output_token);
    else
	_tsasl_buffer_zero(out);

    return error;
}


int
_tsasl_gss_auth(struct tsasl_peer *peer, int flags, struct tsasl_buffer *in,
		struct tsasl_buffer *out)
{
    OM_uint32 maj_stat, min_stat;
    struct tsasl_gss *tg;
    int error;

    tg = peer->tp_mechprivate;

    if (in == NULL && out == NULL) {
	free(tg);
	peer->tp_mechprivate = NULL;
	return 0;
    }

    if (tg == NULL) {
	if ((tg = malloc(sizeof(*tg))) == NULL)
	    return ENOMEM;
	memset(tg, 0, sizeof(*tg));
	tg->tg_flags = 0;
	tg->tg_state = 0;
	tg->tg_hdl = GSS_C_NO_CONTEXT;
	tg->tg_my_max_size = 0xffff;
	tg->tg_peer_max_size = 0;
	tg->tg_neg_sec_bits = 0;
	_tsasl_buffer_zero(&tg->tg_in);
	tg->tg_pkt_off = 0;

	if (peer->tp_flags & TSASL_FLAGS_INITIATOR) {
	    gss_buffer_desc name_buffer;
	    char *name;

	    name_buffer.length = asprintf(&name, "%s@%s", 
					  peer->tp_service,
					  peer->tp_hostname);
	    if (name == NULL) {
		free(tg);
		return ENOMEM;
	    }
	    name_buffer.value = name;

	    maj_stat = gss_import_name(&min_stat,
				       &name_buffer,
				       GSS_C_NT_HOSTBASED_SERVICE,
				       &tg->tg_peer_name);
	    if (GSS_ERROR(maj_stat)) {
		set_gss_errorstring(peer, min_stat, "gss: import name %s",
				    name);
		free(name);
		free(tg);
		return TSASL_ERROR;
	    }
	    free(name);
	} else {
	    tg->tg_peer_name = NULL;
	}
	peer->tp_mechprivate = tg;
    }

    if (peer->tp_flags & TSASL_FLAGS_INITIATOR)
	error = gss_auth_initiator(peer, tg, flags, in, out);
    else
	error = gss_auth_acceptor(peer, tg, flags, in, out);

    return error;
}

int
_tsasl_gss_wrap(struct tsasl_peer *peer, int flags, struct tsasl_buffer *in,
		struct tsasl_buffer *out)
{
    gss_buffer_desc input_token, output_token;
    int error, conf_req_flag, conf_state;
    OM_uint32 maj_stat, min_stat;
    struct tsasl_gss *tg;

    tg = peer->tp_mechprivate;
	
    if ((peer->tp_ctxflags & (TP_CTX_FLAGS_CONF|TP_CTX_FLAGS_INT)) == 0) {
	error = _tsasl_buffer_alloc(out, in->tb_size);
	if (error == 0)
	    memcpy(out->tb_data, in->tb_data, in->tb_size);
	return error;
    }
    _tsasl_buffer_zero(out);
	
    if (peer->tp_ctxflags & TP_CTX_FLAGS_CONF)
	conf_req_flag = 1;
    else if (peer->tp_ctxflags & TP_CTX_FLAGS_INT)
	conf_req_flag = 0;
    else
	conf_req_flag = 0;

    output_token.length = 0;
	
    if (flags) {
	input_token.value = in->tb_data;
	input_token.length = in->tb_size;

	maj_stat = gss_wrap(&min_stat, 
			    tg->tg_hdl,
			    conf_req_flag,
			    GSS_C_QOP_DEFAULT,
			    &input_token,
			    &conf_state,
			    &output_token);
	if (GSS_ERROR(maj_stat)) {
	    set_gss_errorstring(peer, min_stat, "gss: wrap message");
	    error = EINVAL;
	    goto out;
	}

	error = _tsasl_buffer_alloc(out, output_token.length + 4);
	if (error == 0) {
	    unsigned char *p = out->tb_data;

	    p[0] = (output_token.length >> 24) & 0xff;
	    p[1] = (output_token.length >> 16) & 0xff;
	    p[2] = (output_token.length >> 8) & 0xff;
	    p[3] = (output_token.length) & 0xff;
	    memcpy(p + 4, output_token.value, output_token.length);
	}
	if (output_token.length)
	    gss_release_buffer(&min_stat, &output_token);

    } else {
	gss_qop_t qop_state;
	uint32_t pktsz;
	size_t hdrleft;
	unsigned char *p;
	struct tsasl_buffer in_copy, out2;
		
	in_copy = *in;
	_tsasl_buffer_zero(&out2);
		
    again:
	p = in_copy.tb_data;
	hdrleft = 4 - _tsasl_min(4, tg->tg_pkt_off);
	if (hdrleft) {
	    size_t dlen = _tsasl_min(in_copy.tb_size, hdrleft);
			
	    memcpy(&tg->tg_hdr[4 - hdrleft], p, dlen);
	    tg->tg_pkt_off += dlen;
	    if (tg->tg_pkt_off < 4)
		return 0;
	    in_copy.tb_size -= dlen;
	    p += dlen;
			
	    if (_tsasl_parse_pkt_hdr(tg->tg_hdr, &pktsz)) {
		tsasl_buffer_free(out);
		_tsasl_buffer_zero(&tg->tg_in);
		return EINVAL;
	    }
	    if (pktsz > tg->tg_my_max_size) {
		tsasl_buffer_free(out);
		_tsasl_buffer_zero(&tg->tg_in);
		return EINVAL;
	    }
	    error = _tsasl_buffer_alloc(&tg->tg_in, pktsz + 4);
	    if (error)
		return error;
	    memcpy(tg->tg_in.tb_data, tg->tg_hdr, 4);
	    tg->tg_pkt_off = 4;
			
	    if (in_copy.tb_size == 0)
		return 0;
	}
		
	hdrleft = _tsasl_min(tg->tg_in.tb_size - tg->tg_pkt_off, 
			     in_copy.tb_size);
		
	memcpy((char *)tg->tg_in.tb_data + tg->tg_pkt_off,
	       p, 
	       hdrleft);
	p += hdrleft;
	in_copy.tb_data = p;
	in_copy.tb_size -= hdrleft;
		
	tg->tg_pkt_off += hdrleft;
		
	if (tg->tg_pkt_off < tg->tg_in.tb_size)
	    return 0;
		
	tg->tg_pkt_off = 0;
		
	input_token.value = (char *)tg->tg_in.tb_data + 4;
	input_token.length = tg->tg_in.tb_size - 4;
		
	maj_stat = gss_unwrap(&min_stat,
			      tg->tg_hdl,
			      &input_token,
			      &output_token,
			      &conf_state,
			      &qop_state);
	tsasl_buffer_free(&tg->tg_in);
	if (GSS_ERROR(maj_stat)) {
	    set_gss_errorstring(peer, min_stat,
				"gss: unwrap message");
	    tsasl_buffer_free(out);
	    return TSASL_ERROR;
	}

	if ((!conf_state && conf_req_flag) ||
	    (conf_state && !conf_req_flag)) {
	    _tsasl_set_errorstring(peer, "gss: conf state and "
				   "req conf state disagree");
	    tsasl_buffer_free(out);
	    gss_release_buffer(&min_stat, &output_token);
	    return TSASL_ERROR;
	}
		
	error = _tsasl_buffer_alloc(out, 
				    output_token.length+out2.tb_size);
	if (error == 0) {
	    memcpy(out->tb_data, out2.tb_data, out2.tb_size);
	    memcpy((char *)out->tb_data + out2.tb_size,
		   output_token.value, 
		   output_token.length);
	    tsasl_buffer_free(&out2);
	    out2 = *out;
	} else
	    tsasl_buffer_free(&out2);
		
	if (output_token.length)
	    gss_release_buffer(&min_stat, &output_token);
		
	if (error == 0 && in_copy.tb_size != 0)
	    goto again;
    }
		
 out:
    if (GSS_ERROR(maj_stat)) {
	set_gss_errorstring(peer, min_stat, "gss: wrap/unwrap failed");
	return TSASL_ERROR;
    }

    return error;
}

int
_tsasl_gss_status(struct tsasl_peer *peer, char **status)
{
    struct tsasl_gss *tg;
    OM_uint32 maj_stat, min_stat;
    gss_name_t src_name, targ_name;
    gss_buffer_desc src_buf, targ_buf;

    src_buf.length = targ_buf.length = 0;

    tg = peer->tp_mechprivate;
    
    maj_stat = gss_inquire_context(&min_stat, tg->tg_hdl,
				   &src_name, &targ_name,
				   NULL, NULL, NULL, NULL, NULL);
    if (GSS_ERROR(maj_stat)) {
	/* XXX */
	asprintf(status, "gss inquire context error: %d/%d", 
		 maj_stat, min_stat);
	return 0;
    }

    maj_stat = gss_display_name(&min_stat, src_name, 
				&src_buf, NULL);
    if (GSS_ERROR(maj_stat))
	goto out;

    maj_stat = gss_display_name(&min_stat, targ_name, 
				&targ_buf, NULL);
    if (GSS_ERROR(maj_stat))
	goto out;

    asprintf(status, "target: %.*s source: %.*s",
	     (int)targ_buf.length, (char *)targ_buf.value,
	     (int)src_buf.length, (char *)src_buf.value);

 out:
    if (src_buf.length != 0)
	gss_release_buffer(&min_stat, &src_buf);
    if (src_buf.length != 0)
	gss_release_buffer(&min_stat, &targ_buf);

    gss_release_name(&min_stat, &src_name);
    gss_release_name(&min_stat, &targ_name);

    return 0;
}

