/*
 * Copyright (c) 2003 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden). 
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 */

#include "tsasl_locl.h"

RCSID("$Id$");

#include <getarg.h>
#include <roken.h>
#include <vers.h>

typedef void (*wait_tag_func)(void *, struct tsasl_prot *, char *);

static int version_flag = 0;
static int help_flag	= 0;
static int verbose_flag	= 0;
static int pre_login_verbose_flag = 1;

static char *capability;
static char *error_msg;
static int auth_ok;

static struct getargs args[] = {
    {"verbose",	0,	arg_flag,	&verbose_flag,
     "verbose debuging", NULL },
    {"pre-login-verbose",0,	arg_negative_flag, &pre_login_verbose_flag,
     "pre login verbose debuging", NULL },
    {"version",	0,	arg_flag,	&version_flag,
     "print version", NULL },
    {"help",	0,	arg_flag,	&help_flag,
     NULL, NULL }
};

static void
seterror(const char *fmt, ...)
{
    va_list va;

    if (error_msg)
	free(error_msg);
    va_start(va, fmt);
    vasprintf(&error_msg, fmt, va);
    va_end(va);
}

static void
clearerror(void)
{
    free(error_msg);
    error_msg = NULL;
}

static int
connect_host(const char *host, const char *port)
{
    struct addrinfo hints, *res, *res0;
    int error, s;
	
    hints.ai_family = PF_UNSPEC;
    hints.ai_socktype = SOCK_STREAM;
    hints.ai_flags = 0;
    hints.ai_protocol = 0;
    hints.ai_addrlen = 0;
    hints.ai_canonname = NULL;
    hints.ai_addr = NULL;
    hints.ai_next = NULL;

    error = getaddrinfo(host, port, &hints, &res0);
    if (error)
	errx(1, "getaddrinfo: %s", gai_strerror(error));

    s = -1;
    seterror("no address found");
    for (res = res0; res; res = res->ai_next) {
	s = socket(res->ai_family, res->ai_socktype,
		   res->ai_protocol);
	if (s < 0) {
	    seterror("socket: %s (%d)", strerror(errno), errno);
	    continue;
	}
	if (connect(s, res->ai_addr, res->ai_addrlen) < 0) {
	    seterror("connect: %s (%d)", strerror(errno), errno);
	    close(s);
	    s = -1;
	    continue;
	}
	break;
    }
    if (s < 0)
	errx(1, "%s", error_msg);

    clearerror();
    freeaddrinfo(res0);

    return s;
}

/* XXX save state and read more the one char */
static void
getline(struct tsasl_prot *prot, char **line)
{
    int i, len, ret;
    char *l, *p;

    *line = l = NULL;
    i = len = 0;
	
    while (1) {
	if (len == i) {
	    len = (len + 1) * 2;
	    if (len > 64 * 1024)
		errx(1, "excessive answer length");
	    l = erealloc(l, len);
	}
	ret = tsasl_read(prot, 0, &l[i], 1); /* XXX */
	if (ret == TSASL_PROT_EOF)
	    errx(1, "server closed conn");
	i += ret;
	l[i] = '\0';
	if ((p = strstr(l, "\r\n")) != NULL) {
	    *p = '\0';
	    break;
	}
    }
    *line = l;
}

static void
sendmessage(struct tsasl_prot *conn, struct tsasl_peer *peer, 
	    char **tag, const char *fmt, ...)
{
    static int msgnum = 0;
    int len;
    va_list va;
    char *str, *str2;

    va_start(va, fmt);
    vasprintf(&str2, fmt, va);
    if (str2 == NULL)
	errx(1, "out of memory");
    va_end(va);

    if (tag) {
	asprintf(tag, "A%03d", msgnum++);
	if (*tag == NULL)
	    errx(1, "out of memory");
    }

    if (pre_login_verbose_flag)
	printf("%s%s%.*s\n", 
	       tag ? *tag : "",
	       tag ? " " : "",
	       50,
	       str2);

    len = asprintf(&str, "%s%s%s\r\n",
		   tag ? *tag : "",
		   tag ? " " : "",
		   str2);
    free(str2);
    if (str == NULL)
	errx(1, "out of memory");

    tsasl_write(conn, TSASL_PROT_FLUSH, str, len);
    free(str);
}

static void
wait_tag(struct tsasl_prot *conn, const char *tag,
	 wait_tag_func func, void *func_ctx)
{
    int len;
    char *line;

    len = strlen(tag);

    while (1) {
	getline(conn, &line);

	if (pre_login_verbose_flag)
	    printf("%s\n", line);

	if (strncmp(tag, line, len) == 0)
	    break;

	if (line[0] == '*' && line[1] == ' ') {
	    if (strncmp(&line[2], "CAPABILITY", 10) == 0)
		capability = strdup(line);
	    else if (verbose_flag)
		printf("ignoring unknown message: %50s\n",
		       &line[2]);
	} else if (line[0] == '+' && line[1] == ' ') {
	    if (func == NULL)
		errx(1, "got continuation (+) without func");
	    (*func)(func_ctx, conn, line);
	}
	free(line);

    }
    if (line[len] != ' ')
	errx(1, "tag didn't end where we expected");
    if (strncmp(&line[len + 1], "BAD", 3) == 0)
	errx(1, "something BAD: %s\n", &line[len + 1 + 3]);
    else if (strncmp(&line[len + 1], "FAIL", 4) == 0)
	errx(1, "something FAIL: %s\n", &line[len + 1 + 4]);
    if (strncmp(&line[len + 1], "OK", 2) == 0)
	;
    else
	errx(1, "something unknown: %s\n", &line[len + 1]);
    free(line);
}

static void
auth_cont(void *ctx, struct tsasl_prot *conn, char *str)
{
    struct tsasl_peer *client;
    struct tsasl_buffer in, out;
    int error;

    client = ctx;

    in.tb_data = str + 2;
    in.tb_size = strlen(str) - 2;
    if (in.tb_size < 0)
	in.tb_size = 0;
    out.tb_size = 0;

    error = tsasl_request(client, &in, &out);

    if (error != TSASL_CONTINUE && out.tb_size == 0) {

    } else {
	sendmessage(conn, NULL, NULL, "%.*s",out.tb_size, out.tb_data);
	tsasl_buffer_free(&out);
    }

    if (error == TSASL_ERROR) {
	errx(1, "sasl error: %s\n", tsasl_get_errorstring(client));
    } else if (error == TSASL_CONTINUE) {
	/* abort(); */
    } else if (error == TSASL_DONE) {
	auth_ok = 1;
    } else {
	errx(1, "unknown error %d\n", error);
    }
}


static void
usage (int ret)
{
    arg_printusage (args,
		    sizeof(args)/sizeof(*args),
		    NULL,
		    "host");
    exit (ret);
}

int
main(int argc, char **argv)
{
    struct tsasl_peer *client;
    char *line, *tag, *host;
    const char *mech;
    int s, flags, ret;
    int optind = 0;
    struct tsasl_prot *conn;

    setprogname(argv[0]);
	
    if(getarg(args, sizeof(args) / sizeof(args[0]), argc, argv, &optind))
	usage(1);
	
    if (help_flag)
	usage(0);
	
    if(version_flag){
	print_version(NULL);
	exit(0);
    }
	
    argc -= optind;
    argv += optind;

    if (argc != 1)
	usage(1);

    host = argv[0];

    flags = TSASL_FLAGS_INITIATOR |
	TSASL_FLAGS_BASE64_AUTH |
	TSASL_FLAGS_INTEGRITY;

    if (tsasl_peer_init(flags, "imap", host, &client))
	errx(1, "peer init");

    s = connect_host(host, "imap");

    if ((ret = tsasl_prot_init(s, &conn)) != 0)
	errx(1, "tsasl_prot_init failed with %d", ret);

    getline(conn, &line);
    if (verbose_flag)
	printf("s: %.*s\n", 50, line);
    free(line);

    sendmessage(conn, NULL, &tag, "CAPABILITY");
    wait_tag(conn, tag, NULL, NULL);
    free(tag);

    if (capability == NULL)
	errx(1, "server didn't send capability string");

    if (tsasl_find_best_mech(client, capability, &mech, "AUTH=", NULL," "))
	errx(1, "seleting mech from authlist failed");
	
    if (verbose_flag)
	printf("choose mech: %s\n", mech);
    tsasl_select_mech(client, mech);

    sendmessage(conn, NULL, &tag, "AUTHENTICATE %s", mech);
    wait_tag(conn, tag, auth_cont, client);
    free(tag);

    if (auth_ok == 0)
	errx(1, "authentication failed\n");

    tsasl_prot_peer(conn, client);

    /* check capability is as before now that we have integrity */
    {
	char *saved_capability;

	saved_capability = capability;
	capability = NULL;

	sendmessage(conn, client, &tag, "CAPABILITY");
	wait_tag(conn, tag, NULL, NULL);
	free(tag);
		
	if (capability == NULL)
	    errx(1, "server didn't send capability string");
		
	if (strcmp(saved_capability, capability) != 0)
	    errx(1, "CAPABILITY not same after auth");

	free(saved_capability);
    }

    sendmessage(conn, client, &tag, "NOOP");
    wait_tag(conn, tag, NULL, client);
    free(tag);

    while(1) {
	fd_set fds;
	int ret;

	FD_ZERO(&fds);
	FD_SET(s, &fds);
	FD_SET(STDIN_FILENO, &fds);

	ret = select(s + 1, &fds, NULL, NULL, NULL);
	if (ret < 0) {
	    if (errno == EINTR)
		continue;
	    errx(1, "select failed");
	} else if (ret == 0)
	    abort();
				
	if (FD_ISSET(s, &fds)) {
	    char buffer[1000];
	    ssize_t ret;

	    ret = tsasl_read(conn, 0, buffer, sizeof(buffer));
	    if (ret == TSASL_PROT_EOF)
		break;
	    else if (ret < 0)
		errx(1, "unknown error");
	    printf("%.*s", ret, buffer);
	}
	if (FD_ISSET(STDIN_FILENO, &fds)) {
	    char buf[1024];
	    int len;

	    fgets(buf, sizeof(buf), stdin);
	    buf[strcspn(buf, "\n")] = '\0';
	    len = asprintf(&line, "%s\r\n", buf);
	    tsasl_write(conn, TSASL_PROT_FLUSH, line, len);
	    free(line);
	    if (verbose_flag)
		printf("sent: %s", buf);
	}
    }
	
    tsasl_prot_close(conn);
    close(s);

    return 0;
}
