/*
 * Copyright (c) 2003 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden). 
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 */

/*
 * Conf not supported since its downgradeable to an attacker.
 */

#include "tsasl_locl.h"

RCSID("$Id$");

struct md5_value {
    char		*mv_name;
    char		*mv_value;
    struct md5_value *mv_next;
};


#ifdef __BORLANDC__
#include "m_md5.h"
#else
#include <openssl/evp.h>
#include <openssl/rand.h>
#endif
#include <assert.h>

struct tsasl_md5;

static int
_calc_auth(char **, unsigned int *, const char *, unsigned int,
	   const char *, int, struct tsasl_md5 *);

/*
 *
 */

struct md5_server {
    const char *realm;
    const char *snonce;
    const char *snonce_count;
    const char *qop;
    const char *stale;
    const char *maxbuf;
    const char *charset;
    const char *algoritm;
    const char **ciphers;
    const char *auth_param;
};

struct md5_client {
    const char *username;
    const char *cnonce;
    const char *cnonce_count;
    const char *qop;
    const char *digest_uri;
    const char *serv_name;
    const char *response;
    const char *cipher;
    const char *authzid;
    const char *rspauth;
};	

struct tsasl_md5 {
    int tm_flags;
#define TM_F_INT	1
#define TM_F_CONF	2
#define TM_F_C_EMPTY	4
    char *tm_sresponse;
    char *tm_snonce;
    char *tm_cnonce;
    char *tm_digest_uri;
    char *tm_realm;
    unsigned char  tm_int_key_s2c[EVP_MAX_MD_SIZE];
    unsigned char  tm_int_key_c2s[EVP_MAX_MD_SIZE];
    uint32_t tm_int_seq_m2i;
    uint32_t tm_int_seq_i2m;
    /* tm_client and tm_server contains pointers to memory
       mananged by other parts of the code */
    struct md5_client tm_client;
    struct md5_server tm_server;
    const char *tm_need_qop;
};

static char *int_key_c2s =
"Digest session key to client-to-server signing key magic constant";

static char *int_key_s2c =
"Digest session key to server-to-client signing key magic constant";
/*
 *
 */

void
_tsasl_md5_free_value(struct md5_value *val)
{
    struct md5_value *v;
    while(val) {
	v = val->mv_next;
	if (val->mv_name)
	    free(val->mv_name);
	if (val->mv_value)
	    free(val->mv_value);
	free(val);
	val = v;
    }
}

char *
_tsasl_md5_value_find(struct md5_value *val, const char *v)
{
    while (val) {
	if (strcasecmp(v, val->mv_name) == 0)
	    break;
	val = val->mv_next;
    }
    if (val == NULL)
	return NULL;
    return val->mv_value;
}

int
_tsasl_md5_parse_value(struct tsasl_buffer *buf, struct md5_value **val)
{
    struct md5_value *v;
    char *str, *p1, *p2;
    size_t sz;

    *val = NULL;

    if ((str = malloc(buf->tb_size + 1)) == NULL)
	return ENOMEM;
    memcpy(str, buf->tb_data, buf->tb_size);
    str[buf->tb_size] = '\0';

    p1 = str;

    while (p1 - str < buf->tb_size) {
	sz = strspn(p1, " \t\n\r,");
	if (p1[sz] == '\0')
	    break;
	p1 += sz;
	sz = strcspn(p1, " \t\n\r=");
	if (sz == 0 || p1[sz] == '\0')
	    goto error;
	p2 = p1 + sz;

	if ((v = malloc(sizeof(*v))) == NULL)
	    goto nomem;
	v->mv_name = v->mv_value = NULL;
	v->mv_next = *val;
	*val = v;
	if ((v->mv_name = malloc(p2 - p1 + 1)) == NULL)
	    goto nomem;
	strncpy(v->mv_name, p1, p2 - p1);
	v->mv_name[p2 - p1] = '\0';

	sz = strspn(p2, " \t\n\r");
	if (p2[sz] == '\0')
	    goto error;
	p2 += sz;
	assert(*p2 == '=');

	p2++;

	sz = strspn(p2, " \t\n\r");
	if (p2[sz] == '\0')
	    goto error;
	p2 += sz;
	p1 = p2;
		
	if (*p2 == '"') {
	    p1++;
	    while (*p2 == '"') {
		p2++;
		p2 = strchr(p2, '\"');
		if (p2 == NULL)
		    goto error;
		if (p2[0] == '\0')
		    goto error;
		if (p2[-1] != '\\')
		    break;
	    }
	} else {
	    sz = strcspn(p2, " \t\n\r=,");
	    p2 += sz;
	}

#if 0 /* allow empty values */
	if (p1 == p2)
	    goto error;
#endif

	if ((v->mv_value = malloc(p2 - p1 + 1)) == NULL)
	    goto nomem;
	strncpy(v->mv_value, p1, p2 - p1);
	v->mv_value[p2 - p1] = '\0';
		
	if (p2[0] == '\0')
	    break;
	if (p2[0] == '"')
	    p2++;

	sz = strspn(p2, " \t\n\r");
	if (p2[sz] == '\0')
	    break;
	p2 += sz;

	if (p2[0] == '\0')
	    break;
	if (p2[0] != ',')
	    goto error;
	p1 = p2;
    }

    return 0;
 error:
    _tsasl_md5_free_value(*val);
    *val = NULL;
    free(str);
    return EINVAL;
 nomem:
    _tsasl_md5_free_value(*val);
    *val = NULL;
    free(str);
    return ENOMEM;
}

/*
 *
 */

static char *hexchar = "0123456789abcdef";

static int
hex_encode(const void *data, int size, char **str)
{
    const unsigned char *f;
    unsigned char *s;
    int newsize;

    newsize = size * 2;
    *str = NULL;

    if ((s = malloc(newsize + 1)) == NULL)
	return -1;
    *str = (char *)s;
    f = data;
    while(size--) {
	*s++ = hexchar[(*f & 0xf0) >> 4];
	*s++ = hexchar[*f++ & 0xf];
    }
    *s = '\0';
    return newsize;
}


static int
_calc_auth(char **d, unsigned int *size, const char *hA1, unsigned int sizeA1,
	   const char *hA2, int sizeA2, struct tsasl_md5 *tm)
{
    unsigned char digest[EVP_MAX_MD_SIZE];
    EVP_MD_CTX ctx;
    unsigned int dsize;

    EVP_DigestInit(&ctx, EVP_md5());
    EVP_DigestUpdate(&ctx, hA1, sizeA1);
    EVP_DigestUpdate(&ctx, ":", 1);
    EVP_DigestUpdate(&ctx, tm->tm_server.snonce, 
		     strlen(tm->tm_server.snonce));
    EVP_DigestUpdate(&ctx, ":", 1);
    EVP_DigestUpdate(&ctx, tm->tm_client.cnonce_count, 
		     strlen(tm->tm_client.cnonce_count));
    EVP_DigestUpdate(&ctx, ":", 1);
    EVP_DigestUpdate(&ctx, tm->tm_client.cnonce, 
		     strlen(tm->tm_client.cnonce));
    EVP_DigestUpdate(&ctx, ":", 1);
    EVP_DigestUpdate(&ctx, tm->tm_client.qop, 
		     strlen(tm->tm_client.qop));
    EVP_DigestUpdate(&ctx, ":", 1);
    EVP_DigestUpdate(&ctx, hA2, sizeA2);

    EVP_DigestFinal(&ctx, digest, &dsize);

    *size = hex_encode(digest, dsize, d);
    if (*size < 0)
	return ENOMEM;
    return 0;
}

static char conf_zeros[] = ":00000000000000000000000000000000";

static void
free_md5_auth(struct tsasl_md5 *tm)
{
    if (tm) {
	if (tm->tm_sresponse)
	    free(tm->tm_sresponse);
	if (tm->tm_snonce)
	    free(tm->tm_snonce);
	if (tm->tm_cnonce)
	    free(tm->tm_cnonce);
	if (tm->tm_digest_uri)
	    free(tm->tm_digest_uri);
	if (tm->tm_realm)
	    free(tm->tm_realm);
	free(tm);
    }
}

/* 
 * Choose a qop from the list, take the "best one" in this order
 * (auth-conf), auth-int, auth
 */

const char *
_tsasl_md5_choose_qop(const char *qop_string)
{
    const char *p1, *p2;
    int len, sz;
    char *qop;

    len = strlen(qop_string);
    qop = NULL;

    p1 = qop_string;
    while (len > p1 - qop_string) {

	sz = strspn(p1, " \t\n\r,");
	if (p1[sz] == '\0')
	    break;
	p1 += sz;
	if (p1[0] == ',')
	    continue;
	sz = strcspn(p1, " \t\n\r,");
	if (sz == 0)
	    break;
	p2 = p1 + sz;
	if (p2 - p1 == 8 && strncmp(p1, "auth-int", 8) == 0) {
	    if (qop == NULL || strlen(qop) < 8)
		qop = "auth-int";
	}
	if (p2 - p1 == 4 && strncmp(p1, "auth", 4) == 0)
	    if (qop == NULL)
		qop = "auth";
	p1 = p2;
    }
    return qop;
}

int
_tsasl_md5_auth(struct tsasl_peer *peer, int flags, struct tsasl_buffer *in,
		struct tsasl_buffer *out)
{
    struct tsasl_md5 *tm;
    EVP_MD_CTX ctx;
    char password[100];
    unsigned char digest[EVP_MAX_MD_SIZE];
    unsigned char digestA1[EVP_MAX_MD_SIZE];
    char *HdigestA1;
    unsigned char digestA2[EVP_MAX_MD_SIZE];
    char *HdigestA2;
    unsigned int size, sizea1, sizeha1, sizea2;
    char *cresponse, *sresponse;
    struct md5_value *val;

    tm = peer->tp_mechprivate;

    if (in == NULL && out == NULL) {
	free_md5_auth(tm);
	peer->tp_mechprivate = NULL;
	return 0;
    }

    if (tm == NULL) {
	char *s;
	unsigned int n1, n2;

	if (RAND_bytes((char *)&n1, sizeof(n1)) != 1 ||
	    RAND_bytes((char *)&n2, sizeof(n2)) != 1)
	{
	    _tsasl_set_errorstring(peer, "rand bytes broken");
	    free_md5_auth(tm);
	    return TSASL_ERROR;
	}

	if (peer->tp_flags & TSASL_FLAGS_CONFIDENTIALITY)
	    return EINVAL;

	if ((tm = malloc(sizeof(*tm))) == NULL)
	    return ENOMEM;
	memset(tm, 0, sizeof(*tm));
	tm->tm_flags = 0;
	tm->tm_sresponse = NULL;

	asprintf(&s, "%08x%08x", n1, n2);
	if (s == NULL) {
	    free_md5_auth(tm);
	    return ENOMEM;
	}

	if (peer->tp_flags & TSASL_FLAGS_INITIATOR)
	    tm->tm_cnonce = s;
	else
	    tm->tm_snonce = s;

	s = peer->tp_servicename;
	asprintf(&tm->tm_digest_uri, "%s/%s%s%s",
		 peer->tp_service,
		 peer->tp_hostname,
		 s ? "/" : "",
		 s ? s : "");

	if (tm->tm_digest_uri == NULL) {
	    free_md5_auth(tm);
	    return ENOMEM;
	}

	if (peer->tp_realm)
	    tm->tm_realm = strdup(peer->tp_realm);
	else
	    asprintf(&tm->tm_realm, "%s/%s", 
		     peer->tp_service, peer->tp_hostname);

	if (tm->tm_realm == NULL) {
	    free_md5_auth(tm);
	    return ENOMEM;
	}

	if (peer->tp_flags & TSASL_FLAGS_INTEGRITY)
	    peer->tp_ctxflags |= TP_CTX_FLAGS_INT;
	tm->tm_need_qop = NULL;
	tm->tm_int_seq_m2i = 0;
	tm->tm_int_seq_i2m = 0;

	peer->tp_mechprivate = tm;
    }

    if (in->tb_size == 0) {
	char *qop;
	char *buf;

	if (peer->tp_flags & TSASL_FLAGS_INITIATOR) {
	    if (tm->tm_sresponse) {
		_tsasl_set_errorstring(peer, "md5: zero input "
				       "after first exchange");
		return TSASL_ERROR;
	    }
	    if (tm->tm_flags & TM_F_C_EMPTY) {
		_tsasl_set_errorstring(peer, "md5: zero input "
				       "in client twice");
		return TSASL_ERROR;
	    }
	    tm->tm_flags |= TM_F_C_EMPTY;
	    _tsasl_buffer_zero(out);
	    return TSASL_CONTINUE;
	}

	if (peer->tp_ctxflags & TP_CTX_FLAGS_INT)
	    qop = "auth-int";
	else
	    qop = "auth,auth-int";

	tm->tm_need_qop = qop;
			
	size = asprintf(&buf, 
			"realm=%s,nonce=\"%s\",qop=\"%s\","
			"algorithm=md5",
			peer->tp_realm, tm->tm_snonce, qop);

	if (buf == NULL) {
	    _tsasl_set_errorstring(peer, "md5: out of mem");
	    return TSASL_ERROR;
	}
		
	out->tb_data = buf;
	out->tb_size = size;
		
	return TSASL_CONTINUE;
    }

    if ((peer->tp_flags & TSASL_FLAGS_INITIATOR) == 0) {
	if (tm->tm_need_qop == NULL) {
	    _tsasl_set_errorstring(peer, 
				   "md5: server didn't get msg");
	    return TSASL_ERROR;
	}
    }

    if (_tsasl_md5_parse_value(in, &val) != 0) {
	_tsasl_set_errorstring(peer, "Failed to parse %s response",
			       peer->tp_flags & TSASL_FLAGS_INITIATOR ?
			       "client" : "server");
	return TSASL_ERROR;
    }

    if (tm->tm_sresponse) {
	char *rspauth;

	rspauth = _tsasl_md5_value_find(val, "rspauth");

	if ((peer->tp_flags & TSASL_FLAGS_INITIATOR) == 0) {
	    _tsasl_set_errorstring(peer, "md5: server response");
	    _tsasl_md5_free_value(val);
	    return TSASL_ERROR;
	}
	if (rspauth == NULL) {
	    _tsasl_set_errorstring(peer, "md5: server "
				   "failed to send rspauth");
	    _tsasl_md5_free_value(val);
	    return TSASL_ERROR;
	}
	if (strcmp(tm->tm_sresponse, rspauth) != 0) {
	    _tsasl_set_errorstring(peer, "md5: server "
				   "sent invalid auth response");
	    _tsasl_md5_free_value(val);
	    return TSASL_ERROR;
	}
	_tsasl_md5_free_value(val);
	free(tm->tm_sresponse);
	tm->tm_sresponse = NULL;
	out->tb_size = 0;
	out->tb_data = NULL;

	peer->tp_ctxflags |= TP_CTX_FLAGS_AUTHDONE;

	return TSASL_DONE;
    }

#define set_var(_var,_varstr,_type,_m)				\
	if (((_var) = _tsasl_md5_value_find(val, _varstr)) == NULL && _m) {\
		_tsasl_md5_free_value(val);				\
		_tsasl_set_errorstring(peer, #_type " missing " #_varstr);\
		return TSASL_ERROR;					\
	}

    if (peer->tp_flags & TSASL_FLAGS_INITIATOR) {
	set_var(tm->tm_server.realm, "realm", server, 1);
	set_var(tm->tm_server.snonce, "nonce", server, 1);
	set_var(tm->tm_server.qop, "qop", server, 1);
		
	tm->tm_client.qop = _tsasl_md5_choose_qop(tm->tm_server.qop);
	if (tm->tm_client.qop == NULL) {
	    _tsasl_set_errorstring(peer, "md5: not valid qop (%s) ",
				   tm->tm_server.qop);
	    _tsasl_md5_free_value(val);
	    return TSASL_ERROR;
	}

	tm->tm_client.username = peer->tp_authentication_id;
	tm->tm_client.cnonce = tm->tm_cnonce;
	tm->tm_client.cnonce_count = "00000001";
	tm->tm_client.digest_uri = tm->tm_digest_uri;
	if (peer->tp_authorization_id)
	    tm->tm_client.authzid = peer->tp_authorization_id;
	else
	    tm->tm_client.authzid = NULL;
    } else {
	set_var(tm->tm_client.username, "username", client, 1);
	set_var(tm->tm_client.cnonce, "cnonce", client, 1);
	set_var(tm->tm_client.cnonce_count, "nc", client, 1);
	set_var(tm->tm_client.digest_uri, "digest-uri", client, 1);
	set_var(tm->tm_client.response, "response", client, 1);
	set_var(tm->tm_client.qop, "qop", client, 1);
	set_var(tm->tm_client.authzid, "authzid", client, 0);

	tm->tm_server.realm = peer->tp_realm;
	tm->tm_server.snonce = tm->tm_snonce;
    }
#undef set_var

    /* check qop */
    {
	const char *qop;
	int qop_val;
		
	qop = tm->tm_client.qop;
	if (strcmp(qop, "auth-conf") == 0) {
	    _tsasl_set_errorstring(peer, 
				   "md5: auth-conf not supported");
	    _tsasl_md5_free_value(val);
	    return TSASL_ERROR;
	} else if (strcmp(qop, "auth-int") == 0) {
	    qop_val = TP_CTX_FLAGS_INT;
	} else if (strcmp(qop, "auth") == 0) {
	    qop_val = 0;
	} else {
	    _tsasl_set_errorstring(peer, "md5: can't get qop (%s) ",
				   qop);
	    _tsasl_md5_free_value(val);
	    return TSASL_ERROR;
	}

	if ((qop_val & peer->tp_ctxflags) != 
	    (peer->tp_ctxflags & (TP_CTX_FLAGS_CONF|TP_CTX_FLAGS_INT)))
	{
	    _tsasl_set_errorstring(peer, "md5: not enough qop (%s)",
				   qop);
	    _tsasl_md5_free_value(val);
	    return TSASL_ERROR;
	}
	peer->tp_ctxflags |= qop_val;
    }

    /* check other values */
    if (peer->tp_flags & TSASL_FLAGS_INITIATOR) {

    } else {
	if (strcmp(tm->tm_digest_uri, tm->tm_client.digest_uri) != 0) {
	    _tsasl_set_errorstring(peer, "md5: client's uri"
				   "mismatch with server's: "
				   "%s != %s",
				   tm->tm_digest_uri,
				   tm->tm_client.digest_uri);
	    _tsasl_md5_free_value(val);
	    return TSASL_ERROR;
	}
    }

    /* get password */
    if (peer->tp_flags & TSASL_FLAGS_INITIATOR) {
	if (_tsasl_password_get(peer, password, sizeof(password))) {
	    _tsasl_md5_free_value(val);
	    return TSASL_ERROR;
	}
    } else {
	if (_tsasl_auth_func("DIGEST-MD5", tm->tm_client.username,
			     tm->tm_digest_uri, 
			     password, sizeof(password),
			     tsasl_default_auth_db_func)) {
	    _tsasl_set_errorstring(peer, "md5: "
				   "%s in %s didn't have a "
				   "password", 
				   tm->tm_client.username,
				   tm->tm_digest_uri);
	    _tsasl_md5_free_value(val);
	    return TSASL_ERROR;
	}
    }

    /* A1 */
    EVP_DigestInit(&ctx, EVP_md5());
    EVP_DigestUpdate(&ctx, tm->tm_client.username, 
		     strlen(tm->tm_client.username));
    EVP_DigestUpdate(&ctx, ":", 1);
    EVP_DigestUpdate(&ctx, tm->tm_server.realm, 
		     strlen(tm->tm_server.realm));
    EVP_DigestUpdate(&ctx, ":", 1);
    EVP_DigestUpdate(&ctx, password, strlen(password));

    EVP_DigestFinal(&ctx, digest, &size);

    EVP_DigestInit(&ctx, EVP_md5());
    EVP_DigestUpdate(&ctx, digest, size);
    EVP_DigestUpdate(&ctx, ":", 1);
    EVP_DigestUpdate(&ctx, tm->tm_server.snonce, 
		     strlen(tm->tm_server.snonce));
    EVP_DigestUpdate(&ctx, ":", 1);
    EVP_DigestUpdate(&ctx, tm->tm_client.cnonce, 
		     strlen(tm->tm_client.cnonce));

    if (tm->tm_client.authzid) {
	EVP_DigestUpdate(&ctx, ":", 1);
	EVP_DigestUpdate(&ctx, tm->tm_client.authzid, 
			 strlen(tm->tm_client.authzid));
    }

    EVP_DigestFinal(&ctx, digestA1, &sizea1);
    /* digestA1 is H(A1) */

    /* A2 */
    EVP_DigestInit(&ctx, EVP_md5());
    EVP_DigestUpdate(&ctx, "AUTHENTICATE:", sizeof("AUTHENTICATE:") - 1);
    EVP_DigestUpdate(&ctx, tm->tm_client.digest_uri, 
		     strlen(tm->tm_client.digest_uri));

    if (tm->tm_flags & (TM_F_INT|TM_F_CONF))
	EVP_DigestUpdate(&ctx, conf_zeros, sizeof(conf_zeros) - 1);
    EVP_DigestFinal(&ctx, digestA2, &sizea2);
    /* digestA2 is H(A2) */
				 
    /* encode them */

    if ((sizeha1 = hex_encode(digestA1, sizea1, &HdigestA1)) < 0) {
	_tsasl_set_errorstring(peer, "md5: out of memory");
	_tsasl_md5_free_value(val);
	return ENOMEM;
    }

    if ((sizea2 = hex_encode(digestA2, sizea2, &HdigestA2)) < 0) {
	free(HdigestA1);
	_tsasl_set_errorstring(peer, "md5: out of memory");
	_tsasl_md5_free_value(val);
	return ENOMEM;
    }

    if (_calc_auth(&cresponse, &size, HdigestA1, sizeha1, 
		   HdigestA2, sizea2, tm)) {
	free(HdigestA1);
	free(HdigestA2);
	_tsasl_md5_free_value(val);
	_tsasl_set_errorstring(peer, "md5: out of memory");
	return ENOMEM;
    }

    free(HdigestA2);

    /* calculate rspauth */ 
	
    EVP_DigestInit(&ctx, EVP_md5());
    EVP_DigestUpdate(&ctx, ":", 1);
    EVP_DigestUpdate(&ctx, tm->tm_client.digest_uri, 
		     strlen(tm->tm_client.digest_uri));
	
    if (tm->tm_flags & (TM_F_INT|TM_F_CONF))
	EVP_DigestUpdate(&ctx, conf_zeros, 
			 sizeof(conf_zeros) - 1);
    EVP_DigestFinal(&ctx, digestA2, &sizea2);
    /* digestA2 is H(A2)rspauth */
	
    if ((sizea2 = hex_encode(digestA2,sizea2,&HdigestA2)) < 0) {
	free(HdigestA1);
	_tsasl_md5_free_value(val);
	_tsasl_set_errorstring(peer, "md5: out of memory");
	return ENOMEM;
    }
	
    if (_calc_auth(&sresponse, &size, HdigestA1, sizeha1, 
		   HdigestA2, sizea2, tm)) {
	free(HdigestA1);
	free(HdigestA2);
	_tsasl_md5_free_value(val);
	_tsasl_set_errorstring(peer, "md5: out of memory");
	return ENOMEM;
    }

    free(HdigestA1);
    free(HdigestA2);

    if (peer->tp_ctxflags & (TP_CTX_FLAGS_CONF|TP_CTX_FLAGS_INT)) {
	unsigned int size;

	EVP_DigestInit(&ctx, EVP_md5());
	EVP_DigestUpdate(&ctx, digestA1, sizeha1);
	EVP_DigestUpdate(&ctx, int_key_s2c, sizeof(int_key_s2c) - 1);
	EVP_DigestFinal(&ctx, tm->tm_int_key_s2c, &size);
	assert (size == 16);

	EVP_DigestInit(&ctx, EVP_md5());
	EVP_DigestUpdate(&ctx, digestA1, sizeha1);
	EVP_DigestUpdate(&ctx, int_key_c2s, sizeof(int_key_c2s) - 1);
	EVP_DigestFinal(&ctx, tm->tm_int_key_c2s, &size);
    }

    if (peer->tp_flags & TSASL_FLAGS_INITIATOR) {
	const char *authzid;
	char *buf;

	authzid = tm->tm_client.authzid;

	size = asprintf(&buf, "username=%s,response=%s,"
			"cnonce=\"%s\",nc=%s,digest-uri=%s,qop=%s"
			"%s%s%s",
			tm->tm_client.username, cresponse,
			tm->tm_client.cnonce,
			tm->tm_client.cnonce_count,
			tm->tm_client.digest_uri,
			tm->tm_client.qop,
			authzid ? "," : "",
			authzid ? "authzid=" : "",
			authzid ? authzid : "");

	free(cresponse);
	_tsasl_md5_free_value(val);

	if (buf == NULL) {
	    _tsasl_set_errorstring(peer, "md5: out of mem");
	    free(sresponse);
	    return TSASL_ERROR;
	}

	out->tb_size = size;
	out->tb_data = buf;
	tm->tm_sresponse = sresponse;
	return TSASL_CONTINUE;
    } else {
	char *buf;

	if (memcmp(cresponse, tm->tm_client.response, size) != 0) {
	    _tsasl_set_errorstring(peer, "md5: client has "
				   "invalid response string");
	    free(sresponse);
	    free(cresponse);
	    _tsasl_md5_free_value(val);
	    return TSASL_ERROR;
	}
	free(cresponse);

	asprintf(&buf, "DIGEST-MD5:%s@%s", 
		 tm->tm_client.username, tm->tm_realm);
	tsasl_authentication_id_set(peer, buf);
	free(buf);

	if (tm->tm_client.authzid)
	    tsasl_authorization_id_set(peer,tm->tm_client.authzid);

	size = asprintf(&buf, "rspauth=%s", sresponse);
		
	free(sresponse);
	_tsasl_md5_free_value(val);

	if (buf == NULL) {
	    _tsasl_set_errorstring(peer, "md5: out of mem");
	    return TSASL_ERROR;
	}

	out->tb_size = size;
	out->tb_data = buf;

	peer->tp_ctxflags |= TP_CTX_FLAGS_AUTHDONE;

	return TSASL_DONE;
    }
}

static void
md5_hmac10(struct tsasl_peer *peer, unsigned char *key,
	   void *in, int sz, unsigned char *hmac_10)
{
    unsigned char ipad[65], opad[65];
    unsigned char digest[EVP_MAX_MD_SIZE];
    EVP_MD_CTX ctx;
    int i;

    /*
     * MD5(K XOR opad, MD5(K XOR ipad, text))
     */

    memset(ipad, 0, sizeof(ipad));
    memset(opad, 0, sizeof(ipad));
    memcpy(ipad, key, 16);
    memcpy(opad, key, 16);

    for (i=0; i<64; i++) {
	ipad[i] ^= 0x36;
	opad[i] ^= 0x5c;
    }

    EVP_DigestInit(&ctx, EVP_md5());
    EVP_DigestUpdate(&ctx, ipad, 64);
    EVP_DigestUpdate(&ctx, in, sz);
    EVP_DigestFinal(&ctx, digest, NULL);

    EVP_DigestInit(&ctx, EVP_md5());
    EVP_DigestUpdate(&ctx, opad, 64);
    EVP_DigestUpdate(&ctx, digest, 16);
    EVP_DigestFinal(&ctx, digest, NULL);

    memcpy(hmac_10, digest, 10);
}


static int
md5_wrap(struct tsasl_peer *peer, int flags, struct tsasl_buffer *in,
	 struct tsasl_buffer *out)
{
    struct tsasl_md5 *tm;
    int len, error;
    unsigned char hmac_10[10], *p, *key;

    tm = peer->tp_mechprivate;

    /* HMAC-10, type(2), seqnum(4), msg */
    len = 10 + 2 + 4 + in->tb_size;

    if ((error = _tsasl_buffer_alloc(out, len)) != 0)
	return error;
	
    if (peer->tp_flags & TSASL_FLAGS_INITIATOR)
	key = tm->tm_int_key_c2s;
    else
	key = tm->tm_int_key_s2c;

    p = (unsigned char *)out->tb_data + 10;
    *p++ = 0x00;
    *p++ = 0x01;
    *p++ = (tm->tm_int_seq_m2i >> 24) & 0xff;
    *p++ = (tm->tm_int_seq_m2i >> 16) & 0xff;
    *p++ = (tm->tm_int_seq_m2i >> 8) & 0xff;
    *p++ = tm->tm_int_seq_m2i & 0xff;
    memcpy(p, in->tb_data, in->tb_size);

    p = (unsigned char *)out->tb_data + 12;

    md5_hmac10(peer, key, p, len - 12, hmac_10);
    memcpy(out->tb_data, hmac_10, 10);

    tm->tm_int_seq_m2i++;

    return 0;
}

static int
md5_unwrap(struct tsasl_peer *peer, int flags, struct tsasl_buffer *in,
	   struct tsasl_buffer *out)
{
    struct tsasl_md5 *tm;
    int len, error;
    uint32_t seqnum;
    unsigned char hmac_10[10], *p, *key;

    tm = peer->tp_mechprivate;

    /* HMAC-10, type(2), seqnum(4), msg */
    if (in->tb_size < 10 + 2 + 4)
	return EINVAL;

    len = in->tb_size - (10 + 2 + 4);

    p = in->tb_data;
    p += 10;
    if (*p++ != 0x00)
	return EINVAL;
    if (*p++ != 0x01)
	return EINVAL;
    seqnum =  *p++ << 24;
    seqnum += *p++ << 16;
    seqnum += *p++ << 8;
    seqnum += *p++;

    if (seqnum != tm->tm_int_seq_m2i)
	return EINVAL;

    if (peer->tp_flags & TSASL_FLAGS_INITIATOR)
	key = tm->tm_int_key_c2s;
    else
	key = tm->tm_int_key_s2c;

    p = (unsigned char *)in->tb_data + 12;

    md5_hmac10(peer, key, p, len + 4, hmac_10);

    if (memcmp((char *)in->tb_data, hmac_10, 10) != 0)
	return EINVAL;

    if ((error = _tsasl_buffer_alloc(out, len)) != 0)
	return error;
    memcpy(out->tb_data, p, len);

    tm->tm_int_seq_m2i++;

    return 0;
}


int
_tsasl_md5_wrap(struct tsasl_peer *peer, int flags, struct tsasl_buffer *in,
		struct tsasl_buffer *out)
{
    int error;

    if (peer->tp_ctxflags & TP_CTX_FLAGS_CONF)
	return EINVAL;
    else if (peer->tp_ctxflags & TP_CTX_FLAGS_INT) {
	if (flags)
	    error = md5_wrap(peer, flags, in, out);
	else
	    error = md5_unwrap(peer, flags, in, out);
    } else {
	error = _tsasl_buffer_alloc(out, in->tb_size);
	if (error == 0)
	    memcpy(out->tb_data, in->tb_data, in->tb_size);
    }

    return error;
}
