/*
 * Copyright (c) 2003 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden). 
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 */

#include "tsasl_locl.h"

RCSID("$Id$");

struct prot_buffer {
    void   *data;
    size_t  size;
    size_t  offset;
};


struct tsasl_prot {
    int			 fd;
    struct tsasl_peer	*peer;
    struct prot_buffer	 in;
    struct prot_buffer	 out;
};

int
TSASL_LIB_FUNCTION
tsasl_prot_init(int fd, struct tsasl_prot **ret_ctx)
{
    struct tsasl_prot *ctx;
    size_t len;
	
    *ret_ctx = NULL;
    len = 10000;
	
    if ((ctx = malloc(sizeof(*ctx))) == NULL)
	return ENOMEM;
    memset(ctx, 0, sizeof(*ctx));
    ctx->in.data = malloc(len);
    ctx->out.data = malloc(len);
    if (ctx->in.data == NULL || ctx->out.data == NULL) {
	free(ctx->in.data);
	free(ctx->out.data);
	free(ctx);
	return ENOMEM;
    }
    ctx->in.size = len;
    ctx->out.size = len;

    ctx->fd = fd;
    *ret_ctx = ctx;
	
    return 0;
}

int
TSASL_LIB_FUNCTION
tsasl_prot_peer(struct tsasl_prot *ctx, struct tsasl_peer *peer)
{
    if ((peer->tp_ctxflags & TP_CTX_FLAGS_AUTHDONE) == 0)
	return EINVAL;
    ctx->peer = peer;
    return 0;
}

static int
flush(struct tsasl_prot *ctx)
{
    ssize_t ret;
    size_t len, off;
    struct tsasl_buffer in, out;
    void *data;

    if (ctx->peer == NULL) {
	data = ctx->out.data;
	len = ctx->out.offset;
	ret = 0;
    } else {
	in.tb_data = ctx->out.data;
	in.tb_size = ctx->out.offset;
	ret = tsasl_wrap(ctx->peer, &in, &out);
	if (ret)
	    abort();
	data = out.tb_data;
	len = out.tb_size;
    }

    ctx->out.offset = 0;

    off = 0;
    while (len) {
	ret = net_write(ctx->fd, (char *)data + off, len - off);
	if (ret == 0) {
	    ret = TSASL_PROT_EOF;
	    break;
	}
	if (ret < 0)
	    break;
	off += ret;
	ret = 0;
    }
    if (ctx->peer)
	tsasl_buffer_free(&out);
    return ret;
}

ssize_t
TSASL_LIB_FUNCTION
tsasl_write(struct tsasl_prot *ctx, int flags, void *data, size_t len)
{
    size_t left, l = len;

    while (l > 0) {
	left = _tsasl_min(ctx->out.size - ctx->out.offset, l);
	memcpy((unsigned char *)ctx->out.data + ctx->out.offset, data, left);
	l -= left;
	ctx->out.offset += left;
	if (ctx->out.offset == ctx->out.size) {
	    int ret;
	    if ((ret = flush(ctx)) != 0)
		return ret;
	}
    }
    if ((flags & TSASL_PROT_FLUSH) && ctx->out.offset)
	flush(ctx);
    return len;
}

ssize_t
TSASL_LIB_FUNCTION
tsasl_flush(struct tsasl_prot *ctx)
{
    return flush(ctx);
}

ssize_t
TSASL_LIB_FUNCTION
tsasl_printf(struct tsasl_prot *ctx, int flags, const char *fmt, ...)
{
    va_list va;
    char *msg;
    ssize_t ret;

    va_start(va, fmt);
    ret = vasprintf(&msg, fmt, va);
    va_end(va);

    if (ret < 0)
	return -2;
    ret = tsasl_write(ctx, flags, msg, ret);
    free(msg);
    return ret;
}


ssize_t
TSASL_LIB_FUNCTION
tsasl_read(struct tsasl_prot *ctx, int flags, void *data, size_t s)
{
    ssize_t ret;
    size_t len;
    int once = 1;

    len = s;

    while (len) {
	if (ctx->in.offset) {
	    size_t l = _tsasl_min(ctx->in.offset, len);
	    memcpy(data, (char *)ctx->in.data, l);
	    len -= l;
	    ctx->in.offset -= l;
	    memmove(ctx->in.data, (char *)ctx->in.data + l, ctx->in.offset);

	    if (len == 0)
		break;
	}

	if (ctx->in.offset != 0)
	    abort();
	
	if (!once)
	    break;
	once = 0;

#ifdef WIN32
	ret = recv(ctx->fd, ctx->in.data, ctx->in.size, 0);
#else
	ret = read(ctx->fd, ctx->in.data, ctx->in.size);
#endif
	if (ret == 0)
	    return TSASL_PROT_EOF;
	if (ret < 0)
	    abort();

	if (ctx->peer) {
	    struct tsasl_buffer in, out;

	    _tsasl_buffer_zero(&out);
	    in.tb_data = ctx->in.data;
	    in.tb_size = ret;
	    
	    ret = tsasl_unwrap(ctx->peer, &in, &out);
	    if (ret == 0 && out.tb_size != 0) {
		memcpy(ctx->in.data, out.tb_data, out.tb_size);
		ctx->in.offset = out.tb_size;
		tsasl_buffer_free(&out);
	    } 
	} else
	    ctx->in.offset = ret;
    }
    return s - len;
}

int
TSASL_LIB_FUNCTION
tsasl_prot_close(struct tsasl_prot *ctx)
{
    flush(ctx);
    free(ctx->in.data);
    free(ctx->out.data);
    free(ctx);
    return 0;
}
