/*
 * Copyright (c) 2003 Kungliga Tekniska Hgskolan
 * (Royal Institute of Technology, Stockholm, Sweden). 
 * All rights reserved. 
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met: 
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright 
 *    notice, this list of conditions and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution. 
 *
 * 3. Neither the name of the Institute nor the names of its contributors 
 *    may be used to endorse or promote products derived from this software 
 *    without specific prior written permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE INSTITUTE AND CONTRIBUTORS ``AS IS'' AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE INSTITUTE OR CONTRIBUTORS BE LIABLE 
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT 
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY 
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF 
 * SUCH DAMAGE. 
 */

#define TSASL_MAIN
#include "tsasl_locl.h"

RCSID("$Id$");

/* `types' are ordered in "best-ness" */

struct tsasl_types {
    tsasl_type	 tst_type;
    char		*tst_name;
    int		 tst_flags;
    tsasl_auth_func	 tst_auth_func;
    tsasl_wrap_func	 tst_wrap_func;
    tsasl_status_func tst_status_func;
} types[] = {
    { TSASL_GSSAPI,		
      "GSSAPI",
      TSASL_FLAGS_CONFIDENTIALITY | TSASL_FLAGS_INTEGRITY,
      _tsasl_gss_auth,
      _tsasl_gss_wrap,
      _tsasl_gss_status
    },
#if 0
    { TSASL_MD5,
      "DIGEST-MD5",
      TSASL_FLAGS_INTEGRITY | TSASL_FLAGS_INTERACTIVE,
      _tsasl_md5_auth,
      _tsasl_md5_wrap
      _tsasl_md5_status
    },
#endif
    { TSASL_ANONYMOUS,
      "ANONYMOUS",
      TSASL_FLAGS_ANONYMOUS,
      _tsasl_anonymous_auth,
      _tsasl_anonymous_wrap,
      _tsasl_anonymous_status
    }
};

static int (*_tsasl_passwd_func)(const char *, char *, size_t) = NULL;
tsasl_auth_db_func _tsasl_auth_func = tsasl_default_auth_db_func;

int
_tsasl_buffer_alloc(struct tsasl_buffer *buf, size_t len)
{
    if ((buf->tb_data = malloc(len)) == NULL) 
	return ENOMEM;
    buf->tb_size = len;
    return 0;
}

static char _tsasl_zero_buffer[] = "";

void
_tsasl_buffer_zero(struct tsasl_buffer *buf)
{
    buf->tb_data = _tsasl_zero_buffer;
    buf->tb_size = 0;
}

int
_tsasl_buffer_realloc(struct tsasl_buffer *buf, size_t len)
{
    void *ptr;

    if ((ptr = realloc(buf->tb_data, len)) == NULL)
	return ENOMEM;
    buf->tb_data = ptr;
    buf->tb_size = len;
    return 0;
}

void
TSASL_LIB_FUNCTION
tsasl_buffer_free(struct tsasl_buffer *buf)
{
    if (buf->tb_data != _tsasl_zero_buffer)
	free(buf->tb_data);
    buf->tb_data = NULL;
    buf->tb_size = 0;
}

int
TSASL_LIB_FUNCTION
tsasl_peer_init(int flags, const char *service, const char *hostname,
		struct tsasl_peer **ret_peer)
{
    struct tsasl_peer *peer;

    if ((flags & TSASL_FLAGS_INITIATOR) == 0 &&
	(flags & TSASL_FLAGS_INTERACTIVE) != 0)
	return EINVAL;

    *ret_peer = NULL;

    if ((peer = malloc(sizeof(*peer))) == NULL)
	return ENOMEM;
    memset(peer, 0, sizeof(*peer));
    peer->tp_flags = flags;
    peer->tp_type = -1;
    peer->tp_ctxflags = 0;
    peer->tp_errstring = NULL;
    if ((peer->tp_service = strdup(service)) == NULL) {
	free(peer);
	return ENOMEM;
    }		
    if ((peer->tp_hostname = strdup(hostname)) == NULL) {
	free(peer->tp_service);
	free(peer);
	return ENOMEM;
    }
    peer->tp_realm = NULL;
    peer->tp_mechprivate = NULL;
	
    *ret_peer = peer;

    return 0;
}

void
TSASL_LIB_FUNCTION
tsasl_peer_free(struct tsasl_peer *peer)
{
    if (peer == NULL)
	return;
    if (peer->tp_type != -1)
	(*types[peer->tp_type].tst_auth_func)(peer, 0, NULL, NULL);
    if (peer->tp_errstring)
	free(peer->tp_errstring);
    if (peer->tp_realm)
	free(peer->tp_realm);
    free(peer->tp_hostname);
    free(peer->tp_service);
    if (peer->tp_servicename)
	free(peer->tp_servicename);
    if (peer->tp_authorization_id)
	free(peer->tp_authorization_id);
    free(peer);
}

const char *
TSASL_LIB_FUNCTION
tsasl_get_errorstring(struct tsasl_peer *peer)
{
    if (peer->tp_errstring == NULL)
	return "Sucesss/no error";
    return peer->tp_errstring;
}

void
_tsasl_set_errorstring(struct tsasl_peer *peer, const char *fmt, ...)
{
    va_list ap;
    if (peer->tp_errstring)
	free(peer->tp_errstring);
    va_start(ap, fmt);
    vasprintf(&peer->tp_errstring, fmt, ap);
    va_end(ap);
}

#define var_set(_name,_structel,_serverok,_clientok)			 \
									 \
int									 \
TSASL_LIB_FUNCTION							 \
tsasl_##_name##_set(struct tsasl_peer *peer, const char *_name)		 \
{									 \
	if (!_clientok && (peer->tp_flags & TSASL_FLAGS_INITIATOR))	 \
		return EPERM;						 \
	if (!_serverok && (peer->tp_flags & TSASL_FLAGS_INITIATOR) == 0) \
		return EPERM;						 \
	if (peer->tp_##_structel)					 \
		free(peer->tp_##_structel);				 \
	if ((peer->tp_##_structel = strdup(_name)) == NULL)		 \
		return ENOMEM;						 \
	return 0;							 \
}									 \
									 \
int									 \
TSASL_LIB_FUNCTION							 \
tsasl_##_name##_get(struct tsasl_peer *peer, char **ret_##_name)	 \
{									 \
	*ret_##_name = NULL;						 \
	if (peer->tp_##_structel == NULL)				 \
		return ENOENT;						 \
	if ((*ret_##_name = strdup(peer->tp_##_structel)) == NULL)	 \
		return ENOMEM;						 \
	return 0;							 \
}									 \
									 \
void									 \
TSASL_LIB_FUNCTION							 \
tsasl_##_name##_free(char * _name)					 \
{									 \
	free(_name);							 \
}

var_set(servicename,servicename,1,1);
var_set(realm,realm,1,0);
var_set(authorization_id,authorization_id,1,1);
var_set(authentication_id,authentication_id,1,1);

#undef var_set

int
TSASL_LIB_FUNCTION
tsasl_authlist_get(struct tsasl_peer *peer, 
		   const char *prefix, const char *suffix, const char *delim,
		   char **auth_list)
{
    int osize, size, i, j, len, plen, slen, dlen;
    int req_flags, neg_req_flags;
    char *al;

    *auth_list = NULL;

    size = plen = slen = 0;

    if (prefix == NULL)
	prefix = "";
    plen = strlen(prefix);

    if (suffix == NULL)
	suffix = "";
    slen = strlen(suffix);

    if (delim == NULL)
	delim = " ";
    dlen = strlen(delim);

    if (peer->tp_flags & TSASL_FLAGS_INITIATOR)
	return EINVAL;

    req_flags = peer->tp_flags & TSASL_FLAGS_REQ_MASK;
    neg_req_flags = peer->tp_flags & TSASL_FLAGS_NEG_REQ_MASK;

    for (i = 0; i < sizeof(types)/sizeof(types[0]); i++) {
	if ((types[i].tst_flags & req_flags) != req_flags)
	    continue;
	if ((types[i].tst_flags & TSASL_FLAGS_NEG_REQ_MASK) &&
	    (types[i].tst_flags & neg_req_flags) == 0)
	    continue;
	size += strlen(types[i].tst_name) + plen + slen + dlen;
    }
    if (size == 0)
	return ENOENT;
	
    if ((al = malloc(size + 1)) == NULL)
	return ENOMEM;
	
    *auth_list = al;

    osize = size;
    for (i = 0, j = 0; i < sizeof(types)/sizeof(types[0]); i++) {
	if ((types[i].tst_flags & req_flags) != req_flags)
	    continue;
	if ((types[i].tst_flags & TSASL_FLAGS_NEG_REQ_MASK) &&
	    (types[i].tst_flags & neg_req_flags) == 0)
	    continue;
	len = snprintf(al, size,
		       "%s%s%s%s", 
		       j ? delim : "", 
		       prefix,
		       types[i].tst_name,
		       suffix);
	if (len < 0) {
	    free(*auth_list);
	    *auth_list = NULL;
	    return ENOMEM;
	}
	j++;
	size -= len;
	al += len;
    }		
	
    return 0;
}

void
TSASL_LIB_FUNCTION
tsasl_authlist_free(char *auth_list)
{
    free(auth_list);
}

static int
_find_mech(struct tsasl_peer *peer, const char *mech, int len)
{
    int i;

    for (i = 0; i < sizeof(types)/sizeof(types[0]); i++) {
	if (strncmp(mech, types[i].tst_name, len) == 0 &&
	    strlen(types[i].tst_name) == len)
	{
	    return i;
	}
    }
    return -1;
}

int
TSASL_LIB_FUNCTION
tsasl_find_best_mech(struct tsasl_peer *peer,
		     const char *mechlist, 
		     const char **mech,
		     const char *prefix, const char *suffix,
		     const char *delim)
{
    const char *p1, *p2;
    int req_flags, neg_req_flags;
    int  len, sz, plen, slen, mechidx, best_idx;
    int allow_interactive;

    *mech = NULL;
    allow_interactive = slen = plen = 0;
    best_idx = INT_MAX;

    req_flags = peer->tp_flags & TSASL_FLAGS_REQ_MASK;
    neg_req_flags = peer->tp_flags & TSASL_FLAGS_NEG_REQ_MASK;

    if (_tsasl_passwd_func != NULL && 
	(peer->tp_flags & TSASL_FLAGS_INITIATOR))
	allow_interactive = 1;

    if (prefix)
	plen = strlen(prefix);
    if (suffix)
	slen = strlen(suffix);
    if (delim == NULL)
	delim = " ";

    len = strlen(mechlist);
    p1 = mechlist;
    while (p1 - mechlist < len) {

	sz = strspn(p1, delim);
	if (p1[sz] == '\0')
	    break;
	p1 += sz;

	if (prefix && strncmp(prefix, p1, plen) != 0)
	    goto next_ws;

	p1 += plen;

	if (suffix) {
	    if ((p2 = strstr(p1, suffix)) == NULL)
		goto next_ws;

	    sz = strcspn(p1, delim);
	    if (p2 - p1 + slen != sz )
		goto next_ws;
	} else {
	    sz = strcspn(p1, delim);
	    p2 = p1 + sz;
	}

	if (p1 == p2)
	    goto next_ws;

	if ((mechidx = _find_mech(peer, p1, p2 - p1)) == -1)
	    goto next_ws;
		
	if ((types[mechidx].tst_flags & TSASL_FLAGS_INTERACTIVE) &&
	    !allow_interactive)
	    goto next_ws;

	if ((types[mechidx].tst_flags & req_flags) != req_flags)
	    goto next_ws;
	if ((types[mechidx].tst_flags & TSASL_FLAGS_NEG_REQ_MASK) &&
	    ((types[mechidx].tst_flags & neg_req_flags) == 0))
	    goto next_ws;

	if (best_idx > mechidx)
	    best_idx = mechidx;

    next_ws:
	mechidx = -1;
	sz = strcspn(p1, delim);
	if (p1[sz] == '\0')
	    break;
	p1 += sz;		
    }
	
    if (best_idx == INT_MAX)
	return ENOENT;

    *mech = types[best_idx].tst_name;

    return 0;
}

int
TSASL_LIB_FUNCTION
tsasl_select_mech(struct tsasl_peer *peer, const char *mech)
{
    int req_flags, neg_req_flags;

    peer->tp_type = _find_mech(peer, mech, strlen(mech));
    if (peer->tp_type == -1)
	return ENOENT;
    req_flags = peer->tp_flags & TSASL_FLAGS_REQ_MASK;
    neg_req_flags = peer->tp_flags & TSASL_FLAGS_NEG_REQ_MASK;

    if ((types[peer->tp_type].tst_flags & req_flags) != req_flags)
	return EPERM;
    if ((types[peer->tp_type].tst_flags & TSASL_FLAGS_NEG_REQ_MASK) &&
	(types[peer->tp_type].tst_flags & neg_req_flags) == 0)
	return EPERM;
    return 0;
}

const char*
TSASL_LIB_FUNCTION
tsasl_get_mech(struct tsasl_peer *peer)
{
    if (peer->tp_type == -1)
	return NULL;
    return types[peer->tp_type].tst_name;
}

int
TSASL_LIB_FUNCTION
tsasl_is_anonymous(struct tsasl_peer *peer, int *bool)
{
    *bool = 0;

    if (peer->tp_type == -1) {
	_tsasl_set_errorstring(peer, "not mech yet");
	return TSASL_ERROR;
    }

    if ((peer->tp_ctxflags & TP_CTX_FLAGS_AUTHDONE) == 0) {
	_tsasl_set_errorstring(peer, "not authenticated yet");
	return TSASL_ERROR;
    }
    if (strcmp(types[peer->tp_type].tst_name, "ANONYMOUS") == 0)
	*bool = 1;
    return 0;
}

int
TSASL_LIB_FUNCTION
tsasl_default_auth_db_func(const char *mech, const char *user, 
			   const char *service, char *mechdata, 
			   size_t mechdatasize, void *def)
{
    mechdata[0] = '\0';

    if (strcmp(mech, "DIGEST-MD5") == 0)
	return ENOENT;
    else if (strcmp(mech, "GSSAPI") == 0)
	return 0;
    else if (strcmp(mech, "ANONYMOUS") == 0)
	return 0;
    else
	return EINVAL;
}

void
TSASL_LIB_FUNCTION
tsasl_auth_db(tsasl_auth_db_func auth_func)
{
    if (auth_func)
	_tsasl_auth_func = auth_func;
}

void
TSASL_LIB_FUNCTION
tsasl_password_func(int (*func)(const char *, char *, size_t))
{
    _tsasl_passwd_func = func;
}

int
_tsasl_password_get(struct tsasl_peer *peer, char *buf, size_t sz)
{
    int error;
    char *uri, *s;
	
    if (_tsasl_passwd_func == NULL)
	return ENOENT;

    s = peer->tp_servicename;
    asprintf(&uri, "%s/%s%s%s",
	     peer->tp_service,
	     peer->tp_hostname,
	     s ? "/" : "",
	     s ? s : "");

    error = (*_tsasl_passwd_func)(uri, buf, sz);
    free(uri);
    return error;
}


int
TSASL_LIB_FUNCTION
tsasl_request(struct tsasl_peer *peer, 
	      struct tsasl_buffer *in, 
	      struct tsasl_buffer *out)
{
    struct tsasl_buffer inc;
    void *inptr;
    int error;

    if (peer->tp_type == -1) {
	_tsasl_set_errorstring(peer, "mech not selected");
	return TSASL_ERROR;
    }


    inc = *in;
    if (inc.tb_size == 0)
	_tsasl_buffer_zero(&inc);
    out->tb_size = 0;
    out->tb_data = NULL;

    if (peer->tp_flags & TSASL_FLAGS_BASE64_AUTH) {
	if ((inptr = malloc((in->tb_size == 0)?1:in->tb_size)) == NULL)
	    return ENOMEM;
	inc.tb_size = base64_decode(inc.tb_data, inptr);
	inc.tb_data = inptr;
	if (inc.tb_size < 0) {
	    free(inptr);
	    return EINVAL;
	}
    } else
	inptr = NULL;

    error = (*types[peer->tp_type].tst_auth_func)(peer, 0, &inc, out);
    if (inc.tb_data == inptr)
	free(inptr);

    if (out->tb_size && (peer->tp_flags & TSASL_FLAGS_BASE64_AUTH)) {
	struct tsasl_buffer outc;
	char *str;
		
	outc.tb_size = base64_encode(out->tb_data, out->tb_size, &str);
	tsasl_buffer_free(out);
	if (outc.tb_size >= 0) {
	    outc.tb_data = str;
	    *out = outc;
	} else
	    error = ENOMEM;
    }
    return error;
}

int
TSASL_LIB_FUNCTION
tsasl_wrap(struct tsasl_peer *peer,
	   struct tsasl_buffer *in, struct tsasl_buffer *out)
{
    struct tsasl_buffer inc;
    int error;

    if ((peer->tp_ctxflags & TP_CTX_FLAGS_AUTHDONE) == 0)
	return EINVAL;

    inc = *in;
    if (inc.tb_size == 0)
	_tsasl_buffer_zero(&inc);

    out->tb_size = 0;

    error = (*types[peer->tp_type].tst_wrap_func)(peer, 1, &inc, out);

    if (out->tb_size && (peer->tp_flags & TSASL_FLAGS_BASE64_WRAP)) {
	struct tsasl_buffer outc;
	char *str;
		
	outc.tb_size = base64_encode(out->tb_data, out->tb_size, &str);
	tsasl_buffer_free(out);
	if (outc.tb_size >= 0) {
	    outc.tb_data = str;
	    *out = outc;
	} else
	    error = ENOMEM;
    }

    return error;
}

int
TSASL_LIB_FUNCTION
tsasl_unwrap(struct tsasl_peer *peer,
	     struct tsasl_buffer *in, struct tsasl_buffer *out)
{
    struct tsasl_buffer inc;
    void *inptr;
    int error;

    if ((peer->tp_ctxflags & TP_CTX_FLAGS_AUTHDONE) == 0)
	return EINVAL;

    inc = *in;
    if (inc.tb_size == 0)
	_tsasl_buffer_zero(&inc);
    out->tb_size = 0;

    if (peer->tp_flags & TSASL_FLAGS_BASE64_WRAP) {
	if ((inptr = malloc(in->tb_size)) == NULL)
	    return ENOMEM;
	inc.tb_size = base64_decode(inc.tb_data, inptr);
	inc.tb_data = inptr;
	if (inc.tb_size < 0) {
	    free(inptr);
	    return EINVAL;
	}
    } else {
	inc = *in;
	inptr = NULL;
    }

    error = (*types[peer->tp_type].tst_wrap_func)(peer, 0, &inc, out);

    if (inptr)
	free(inptr);

    return error;
}

int
TSASL_LIB_FUNCTION
tsasl_status_get(struct tsasl_peer *peer, char **str)
{
    const char *mech;
	
    if (peer->tp_type != -1)
	mech = types[peer->tp_type].tst_name;
    else
	mech = "no mech yet";

    if ((peer->tp_ctxflags & TP_CTX_FLAGS_AUTHDONE) == 0){
	asprintf(str, "auth incomplete (mech %s)", mech);
	return 0;
    } else {
	char *mech_status;

	(*types[peer->tp_type].tst_status_func)(peer, &mech_status);

	asprintf(str, "%s int: %s conf: %s mech: [%s]",
		 mech,  
		 peer->tp_ctxflags & TP_CTX_FLAGS_INT ? "yes" : "no",
		 peer->tp_ctxflags & TP_CTX_FLAGS_CONF ? "yes":"no",
		 mech_status);

	free(mech_status);
    }	

    if (str == NULL)
	return ENOMEM;
    return 0;
}

void
TSASL_LIB_FUNCTION
tsasl_status_free(char *status)
{
    free(status);
}

int
_tsasl_parse_pkt_hdr(unsigned char *p, uint32_t *rlen)
{
    int32_t len;

    len = (p[0] << 24) + (p[1] << 16) + (p[2] << 8) + p[3];
    if (len < 0)
	return EINVAL;

    *rlen = len;
    return 0;
}

size_t
_tsasl_min(size_t a, size_t b)
{
    if (a < b)
	return a;
    return b;
}
